/*
 * Copyright 2006-2007 Red Hat, Inc.
 *
 * Authors:
 *    Peter Jones <pjones@redhat.com>
 *
 * This software may be freely redistributed under the terms of the GNU
 * General Public License, version 2.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#define _GNU_SOURCE 1

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <linux/hdreg.h>

#include <bdevid.h>

struct ata_probe_priv {
    char *vendor;
    char *model;
    char *unique_id;
};

static void ata_probe_destroy_priv(void *privp)
{
    struct ata_probe_priv *priv;
    if (!privp)
        return;
    priv = privp;
    if (priv->model)
        free(priv->model);
    if (priv->unique_id)
        free(priv->unique_id);
    free(priv);
}

static int ata_do_get_model(struct bdevid_bdev *bdev,
        struct ata_probe_priv *priv)
{
    struct hd_driveid hdid;
    int fd;
    char model[41], rev[9];
    int m, n;

    fd = bdevid_bdev_get_fd(bdev);

    if (ioctl(fd, HDIO_GET_IDENTITY, &hdid) < 0)
        return -1;

    memmove(model, hdid.model, 40);
    model[40] = '\0';
    for (m = 0, n = 0; model[m]; m++) {
        if (model[m] != ' ')
            model[n++] = model[m];
    }
    model[n] = '\0';

    memmove(rev, hdid.fw_rev, 8);
    rev[8] = '\0';
    for (m = 0, n = 0; rev[m]; m++) {
        if (rev[m] != ' ')
            rev[n++] = rev[m];
    }
    rev[n] = '\0';

    char dash = '-';
    if (!model[0] && !rev[0])
        return -1;
    else if (!model[0] || !rev[0])
        dash = '\0';

    if (asprintf(&priv->model, "%s%c%s", model[0] ? model : rev, dash, rev) < 0)
        return -1;
    return 0;
}

static int ata_do_get_unique_id(struct bdevid_bdev *bdev,
        struct ata_probe_priv *priv)
{
    struct hd_driveid hdid;
    int fd;
    char serial[21];
    size_t m, n;

    fd = bdevid_bdev_get_fd(bdev);

    memset(&hdid, '\0', sizeof (hdid));
    if (ioctl(fd, HDIO_GET_IDENTITY, &hdid) < 0)
        return -1;

    memset(serial, '\0', sizeof(serial));
    strncpy(serial, (char *)hdid.serial_no, 20);
    for (m = 0, n = 0; serial[m]; m++) {
        if (serial[m] != ' ')
            serial[n++] = serial[m];
    }
    serial[n] = '\0';

    if (!serial[0])
        return -1;
    if (!(priv->unique_id = strdup(serial)))
        return -1;
    return 0;
}

static int ata_do_probe(struct bdevid_bdev *bdev, void **privp)
{
    struct ata_probe_priv *priv;
    if (!privp)
        return -1;
    if (*privp)
        return 0;
    if (!(priv = calloc(1, sizeof (*priv))))
        return -1;

    if (ata_do_get_model(bdev, priv) < 0 || !priv->model ||
            !strlen(priv->model))
        goto err;
    if (ata_do_get_unique_id(bdev, priv) < 0 || !priv->unique_id ||
            !strlen(priv->unique_id))
        goto err;
    priv->vendor = "1ATA";
    *privp = priv;
    return 0;

    int errnum;
err:
    errnum = errno;
    ata_probe_destroy_priv(priv);
    errno = errnum;
    return -1;
}

static int ata_get_unique_id(struct bdevid_bdev *bdev, char **id, void **privp)
{
    if (ata_do_probe(bdev, privp) < 0)
        return -1;
    struct ata_probe_priv *priv = *privp;
    *id = priv->unique_id;
    return 0;
}

static int ata_get_model(struct bdevid_bdev *bdev, char **id, void **privp)
{
    if (ata_do_probe(bdev, privp) < 0)
        return -1;
    struct ata_probe_priv *priv = *privp;
    *id = priv->model;
    return 0;
}

static int ata_get_vendor(struct bdevid_bdev *bdev, char **id, void **privp)
{
    if (ata_do_probe(bdev, privp) < 0)
        return -1;
    struct ata_probe_priv *priv = *privp;
    *id = priv->vendor;
    return 0;
}

static struct bdevid_probe_ops ata_probe_ops = {
    .name = "ata_probe",
    .get_vendor = ata_get_vendor,
    .get_model = ata_get_model,
    .get_unique_id = ata_get_unique_id,
    .destroy_priv = ata_probe_destroy_priv,
};

static int ata_init(struct bdevid_module *bm,
    bdevid_register_func register_probe)
{
    if (register_probe(bm, &ata_probe_ops) == -1)
        return -1;
    return 0;
}

static struct bdevid_module_ops ata_module = {
    .magic = BDEVID_MAGIC,
    .name = "ata",
    .init = ata_init,
};

BDEVID_MODULE(ata_module);

/*
 * vim:ts=8:sw=4:sts=4:et
 */
