/*
 * Copyright 2006-2007 Red Hat, Inc.
 * Copyright 2003 IBM Corp.
 *
 * Authors:
 *    Patrick Mansfield <patmans@us.ibm.com>
 *    Peter Jones <pjones@redhat.com>
 *
 *    This program is free software; you can redistribute it and/or modify it
 *    under the terms of the GNU General Public License as published by the
 *    Free Software Foundation version 2 of the License.
 */ 

#define _GNU_SOURCE 1

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <scsi/sg.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <bdevid.h>

#include "scsi.h"
#include "scsi_id.h"
#include "scsi_id_version.h"

#define info(...)
#define dbg(...)

/*
 * A priority based list of id, naa, and binary/ascii for the identifier
 * descriptor in VPD page 0x83.
 *
 * Brute force search for a match starting with the first value in the
 * following id_search_list. This is not a performance issue, since there
 * is normally one or some small number of descriptors.
 */
static const struct scsi_id_search_values id_search_list[] = {
    { SCSI_ID_NAA,    SCSI_ID_NAA_IEEE_REG_EXTENDED,    SCSI_ID_UTF8 },
    { SCSI_ID_NAA,    SCSI_ID_NAA_IEEE_REG_EXTENDED,    SCSI_ID_ASCII },
    { SCSI_ID_NAA,    SCSI_ID_NAA_IEEE_REG_EXTENDED,    SCSI_ID_BINARY },
    { SCSI_ID_NAA,    SCSI_ID_NAA_IEEE_REG,    SCSI_ID_UTF8 },
    { SCSI_ID_NAA,    SCSI_ID_NAA_IEEE_REG,    SCSI_ID_ASCII },
    { SCSI_ID_NAA,    SCSI_ID_NAA_IEEE_REG,    SCSI_ID_BINARY },
    /*
     * Devices already exist using NAA values that are now marked
     * reserved. These should not conflict with other values, or it is
     * a bug in the device. As long as we find the IEEE extended one
     * first, we really don't care what other ones are used. Using
     * don't care here means that a device that returns multiple
     * non-IEEE descriptors in a random order will get different
     * names.
     */
    { SCSI_ID_NAA,    SCSI_ID_NAA_DONT_CARE,    SCSI_ID_UTF8 },
    { SCSI_ID_NAA,    SCSI_ID_NAA_DONT_CARE,    SCSI_ID_ASCII },
    { SCSI_ID_NAA,    SCSI_ID_NAA_DONT_CARE,    SCSI_ID_BINARY },
    { SCSI_ID_EUI_64,    SCSI_ID_NAA_DONT_CARE,    SCSI_ID_UTF8 },
    { SCSI_ID_EUI_64,    SCSI_ID_NAA_DONT_CARE,    SCSI_ID_ASCII },
    { SCSI_ID_EUI_64,    SCSI_ID_NAA_DONT_CARE,    SCSI_ID_BINARY },
    { SCSI_ID_T10_VENDOR,    SCSI_ID_NAA_DONT_CARE,    SCSI_ID_UTF8 },
    { SCSI_ID_T10_VENDOR,    SCSI_ID_NAA_DONT_CARE,    SCSI_ID_ASCII },
    { SCSI_ID_T10_VENDOR,    SCSI_ID_NAA_DONT_CARE,    SCSI_ID_BINARY },
    { SCSI_ID_VENDOR_SPECIFIC,    SCSI_ID_NAA_DONT_CARE,    SCSI_ID_UTF8 },
    { SCSI_ID_VENDOR_SPECIFIC,    SCSI_ID_NAA_DONT_CARE,    SCSI_ID_ASCII },
    { SCSI_ID_VENDOR_SPECIFIC,    SCSI_ID_NAA_DONT_CARE,    SCSI_ID_BINARY },
};

static const char hex_str[]="0123456789abcdef";

/*
 * Values returned in the result/status, only the ones used by the code
 * are used here.
 */

#define DID_NO_CONNECT            0x01    /* Unable to connect before timeout */
#define DID_BUS_BUSY            0x02    /* Bus remain busy until timeout */
#define DID_TIME_OUT            0x03    /* Timed out for some other reason */
#define DRIVER_TIMEOUT            0x06
#define DRIVER_SENSE            0x08    /* Sense_buffer has been set */

/* The following "category" function returns one of the following */
#define SG_ERR_CAT_CLEAN        0    /* No errors or other information */
#define SG_ERR_CAT_MEDIA_CHANGED    1    /* interpreted from sense buffer */
#define SG_ERR_CAT_RESET        2    /* interpreted from sense buffer */
#define SG_ERR_CAT_TIMEOUT        3
#define SG_ERR_CAT_RECOVERED        4     /* Successful command after recovered err */
#define SG_ERR_CAT_NOTSUPPORTED     5    /* Illegal / unsupported command */
#define SG_ERR_CAT_SENSE        98    /* Something else in the sense buffer */
#define SG_ERR_CAT_OTHER        99    /* Some other error/warning */

static int sg_err_category_new(int scsi_status, int msg_status, int
                   host_status, int driver_status, const
                   unsigned char *sense_buffer, int sb_len)
{
    scsi_status &= 0x7e;

    /*
     * XXX change to return only two values - failed or OK.
     */

    /*
     * checks msg_status
     */
    if (!scsi_status && !msg_status && !host_status && !driver_status)
        return SG_ERR_CAT_CLEAN;

    if ((scsi_status == SCSI_CHECK_CONDITION) ||
        (scsi_status == SCSI_COMMAND_TERMINATED) ||
        ((driver_status & 0xf) == DRIVER_SENSE)) {
        if (sense_buffer && (sb_len > 2)) {
            int sense_key;
            unsigned char asc;

            if (sense_buffer[0] & 0x2) {
                sense_key = sense_buffer[1] & 0xf;
                asc = sense_buffer[2];
            } else {
                sense_key = sense_buffer[2] & 0xf;
                asc = (sb_len > 12) ? sense_buffer[12] : 0;
            }

            if (sense_key == RECOVERED_ERROR)
                return SG_ERR_CAT_RECOVERED;
            else if (sense_key == UNIT_ATTENTION) {
                if (0x28 == asc)
                    return SG_ERR_CAT_MEDIA_CHANGED;
                if (0x29 == asc)
                    return SG_ERR_CAT_RESET;
            } else if (sense_key == ILLEGAL_REQUEST) {
                return SG_ERR_CAT_NOTSUPPORTED;
            }
        }
        return SG_ERR_CAT_SENSE;
    }
    if (!host_status) {
        if ((host_status == DID_NO_CONNECT) ||
            (host_status == DID_BUS_BUSY) ||
            (host_status == DID_TIME_OUT))
            return SG_ERR_CAT_TIMEOUT;
    }
    if (!driver_status) {
        if (driver_status == DRIVER_TIMEOUT)
            return SG_ERR_CAT_TIMEOUT;
    }
    return SG_ERR_CAT_OTHER;
}

static int sg_err_category3(struct sg_io_hdr *hp)
{
    return sg_err_category_new(hp->status, hp->msg_status,
                   hp->host_status, hp->driver_status,
                   hp->sbp, hp->sb_len_wr);
}

static int scsi_inquiry(int fd, unsigned char evpd, unsigned char page,
            unsigned char *buf, size_t buflen)
{
    unsigned char inq_cmd[INQUIRY_CMDLEN] =
        { INQUIRY_CMD, evpd, page, 0, buflen, 0 };
    unsigned char sense[SENSE_BUFF_LEN];
    struct sg_io_hdr io_hdr;
    int retval;
    int retry = 3; /* rather random */

    if (buflen > SCSI_INQ_BUFF_LEN)
        buflen = SCSI_INQ_BUFF_LEN;

    do {
        memset(&io_hdr, '\0', sizeof(struct sg_io_hdr));
        memset(&sense, '\0', sizeof (sense));
        io_hdr.interface_id = 'S';
        io_hdr.cmd_len = sizeof(inq_cmd);
        io_hdr.mx_sb_len = sizeof(sense);
        io_hdr.dxfer_direction = SG_DXFER_FROM_DEV;
        io_hdr.dxfer_len = buflen;
        io_hdr.dxferp = buf;
        io_hdr.cmdp = inq_cmd;
        io_hdr.sbp = sense;
        io_hdr.timeout = DEF_TIMEOUT;

        while ((retval = ioctl(fd, SG_IO, &io_hdr)) < 0)
            if (errno != EAGAIN)
                return retval;

        retval = sg_err_category3(&io_hdr);

        switch (retval) {
            case SG_ERR_CAT_NOTSUPPORTED:
                buf[1] = 0;
                /* Fallthrough */
            case SG_ERR_CAT_CLEAN:
            case SG_ERR_CAT_RECOVERED:
                retval = 0;
                break;
        }

        if (!retval)
            break;

        retval = -1;
    } while (--retry > 0);

    return retval;
}

/* Get list of supported EVPD pages */
static int do_scsi_page0_inquiry(int fd, unsigned char *inq,
        unsigned char *page0, size_t len)
{
    int retval;

    memset(page0, '\0', len);
    if ((retval = scsi_inquiry(fd, 1, 0x0, page0, len) < 0))
        return retval;
    if (page0[1] != 0 || page0[3] > len)
        return -1;

    /* Some ill behaved devices return the standard inquiry instead of evpd,
     * so check against the normal data */

    if (page0[3] >= VENDOR_END) {
       if (!strncmp((char *)&inq[VENDOR_START],
                    (char *)&page0[VENDOR_START], VENDOR_LENGTH))
           return -1;
    }
    if (page0[3] >= MODEL_END) {
       if (!strncmp((char *)&inq[MODEL_START],
                    (char *)&page0[MODEL_START], MODEL_END))
           return -1;
    }
    return 0;
}

struct scsi_probe_priv {
    char *vendor;
    size_t vendor_len;
    char *model;
    size_t model_len;
    char *unique_id_80;
    size_t unique_id_80_len;
    char *unique_id_83;
    size_t unique_id_83_len;
};

/**
 * check_fill_0x83_id - check the page 0x83 id, if OK allocate and fill
 * serial number.
 **/
static int check_fill_0x83_id(unsigned char *page_83, 
    const struct scsi_id_search_values *id_search,
    struct scsi_probe_priv *priv)
{
    int i, j, len;
    char *buf;
    int inpos, outpos, dashpos;

    if (priv->unique_id_83)
        return 0;

    /*
     * ASSOCIATION must be with the device (value 0)
     */
    if ((page_83[1] & 0x30) != 0)
        return 1;

    if ((page_83[1] & 0x0f) != id_search->id_type)
        return 1;

    /*
     * Possibly check NAA sub-type.
     */
    if ((id_search->naa_type != SCSI_ID_NAA_DONT_CARE) &&
            (id_search->naa_type != (page_83[4] & 0xf0) >> 4))
        return 1;

    /*
     * Check for matching code set.
     */
    if ((page_83[0] & 0x0f) != id_search->code_set)
        return 1;

    /*
     * page_83[3]: identifier length
     */
    len = page_83[3];
    switch (page_83[0] & 0x0f) {
        case SCSI_ID_ASCII:
        case SCSI_ID_UTF8:
            break;
        default:
            /*
             * If not ASCII or UTF8, use two bytes for each binary value.
             */
            len *= 2;
    }

    /*
     * Add one byte for the NUL termination.
     */
    len += 1;

    if (!(priv->unique_id_83 = calloc(len, sizeof (char)))) {
        return 1;
    }
    priv->unique_id_83_len = 0;

#if 0
    priv->unique_id_83[0] = hex_str[id_search->id_type];
#endif

    i = 4; /* offset to the start of the identifier */
    j = 0;
    switch (page_83[0] & 0x0f) {
        case SCSI_ID_ASCII:
        case SCSI_ID_UTF8:

            inpos = outpos = 0;
            dashpos = -1;

            if (!(buf = calloc(len, sizeof (char)))) {
                free(priv->unique_id_83);
                priv->unique_id_83 = NULL;
                return 1;
            }

            while (i < (4 + page_83[3]))
                buf[j++] = page_83[i++];
            buf[j] = '\0';

            while (1) {
                int l, m;
                l = sscanf(buf+inpos, " %s%n", priv->unique_id_83+outpos, &m);
                if (l < 0)
                    break;
                if (dashpos > 0)
                    priv->unique_id_83[dashpos] = '-';
                outpos += strlen(priv->unique_id_83+outpos);
                dashpos = outpos;
                priv->unique_id_83[outpos++] = '\0';
                inpos += m;
            }

            free(buf);
            break;
        default:
            /*
             * Binary descriptor, convert to ASCII, using two bytes of
             * ASCII for each byte in the page_83.
             */
            while (i < (4 + page_83[3])) {
                priv->unique_id_83[j++] = hex_str[(page_83[i] & 0xf0) >> 4];
                priv->unique_id_83[j++] = hex_str[ page_83[i] & 0x0f];
                i++;
            }
    }
    return 0;
}

/* Extract the raw binary from VPD 0x83 pre-SPC devices */
static int check_fill_0x83_prespc3(unsigned char *page_83, 
    const struct scsi_id_search_values *id_search, struct scsi_probe_priv *priv)
{
    int i, j;
    
    if (priv->unique_id_83)
        return 0;

    if (!(priv->unique_id_83 = calloc(page_83[3] + 1, sizeof (char))))
        return 1;

#if 0
    priv->unique_id_83[0] = hex_str[id_search->id_type];
#endif
    for (j = 0, i = 0; i < page_83[3]; ++i) {
        if (page_83[4+i] != 0x20) {
            priv->unique_id_83[j++] = hex_str[(page_83[4+i] & 0xf0) >> 4];
            priv->unique_id_83[j++] = hex_str[ page_83[4+i] & 0x0f];
        }
    }
    for (j = 0, i = 0; priv->unique_id_83[j]; j++) {
        if (priv->unique_id_83[j] != ' ')
            priv->unique_id_83[i++] = priv->unique_id_83[j];
    };
    priv->unique_id_83[i] = '\0';
    priv->unique_id_83_len = strlen(priv->unique_id_83);
    if (priv->unique_id_83_len > 1)
        return 0;
    
    free(priv->unique_id_83);
    priv->unique_id_83 = NULL;
    priv->unique_id_83_len = 0;
    return 1;
}

/* Get device identification VPD page */
static int do_scsi_page83_inquiry(int fd, struct scsi_probe_priv *priv)
{
    int retval;
    unsigned int id_ind, j;
    unsigned char page_83[SCSI_INQ_BUFF_LEN];

    memset(page_83, 0, SCSI_INQ_BUFF_LEN);
    if ((retval = scsi_inquiry(fd, 1, PAGE_83, page_83, SCSI_INQ_BUFF_LEN)) < 0)
        return retval;

    if (page_83[1] != PAGE_83)
        return -1;
    
    /*
     * XXX Some devices (IBM 3542) return all spaces for an identifier if
     * the LUN is not actually configured. This leads to identifers of
     * the form: "1            ".
     */

    /*
     * Model 4, 5, and (some) model 6 EMC Symmetrix devices return
     * a page 83 reply according to SCSI-2 format instead of SPC-2/3.
     *
     * The SCSI-2 page 83 format returns an IEEE WWN in binary
     * encoded hexi-decimal in the 16 bytes following the initial
     * 4-byte page 83 reply header.
     *
     * Both the SPC-2 and SPC-3 formats return an IEEE WWN as part
     * of an Identification descriptor.  The 3rd byte of the first
     * Identification descriptor is a reserved (BSZ) byte field.
     *
     * Reference the 7th byte of the page 83 reply to determine
     * whether the reply is compliant with SCSI-2 or SPC-2/3
     * specifications.  A zero value in the 7th byte indicates
     * an SPC-2/3 conformant reply, (i.e., the reserved field of the
     * first Identification descriptor).  This byte will be non-zero
     * for a SCSI-2 conformant page 83 reply from these EMC
     * Symmetrix models since the 7th byte of the reply corresponds
     * to the 4th and 5th nibbles of the 6-byte OUI for EMC, that is,
     * 0x006048.
     */
    
    if (page_83[6] != 0) 
        return check_fill_0x83_prespc3(page_83, id_search_list, priv);

    /*
     * Search for a match in the prioritized id_search_list.
     */
    for (id_ind = 0; id_ind < sizeof(id_search_list)/sizeof(id_search_list[0]);
            id_ind++) {
        /*
         * Examine each descriptor returned. There is normally only
         * one or a small number of descriptors.
         */
        for (j = 4; j <= (unsigned int)page_83[3] + 3; j += page_83[j + 3]+4) {
            retval = check_fill_0x83_id(&page_83[j], &id_search_list[id_ind],
                                        priv);
            if (retval <= 0)
                return retval;
        }
    }
    return 1;
}

/* Get unit serial number VPD page */
static int do_scsi_page80_inquiry(int fd, struct scsi_probe_priv *priv)
{
    int m, n;
    char *s;
    size_t len;
    unsigned char buf[SCSI_INQ_BUFF_LEN];

    if (priv->unique_id_80)
        return 0;

    memset(buf, 0, SCSI_INQ_BUFF_LEN);
    m = scsi_inquiry(fd, 1, PAGE_80, buf, SCSI_INQ_BUFF_LEN);
    if (m < 0)
        return m;

    if (buf[1] != PAGE_80)
        return -1;

    len = buf[3];
    if (!(s = calloc(len, sizeof (char))))
        return -1;

    for (m = 4, n = 0; m < len + 4; m++) {
        if (buf[m] != ' ')
            s[n++] = buf[m];
    }
    buf[m] = '\0';

    priv->unique_id_80_len = strlen(s);
    if (priv->unique_id_80_len) {
        priv->unique_id_80 = s;
        return 0;
    }
    
    free(s);
    return -1;
}

static inline int result_ok(struct scsi_probe_priv *priv, int final)
{
    return final ? (priv->vendor || priv->model ||
                    priv->unique_id_80 || priv->unique_id_83)
                 : (priv->vendor && priv->model &&
                    (priv->unique_id_80 || priv->unique_id_83));
}

static int scsi_get_devattr(struct bdevid_bdev *bdev, char *attr, char **value,
        size_t *len)
{
    char *sysfs_dir = bdevid_bdev_get_sysfs_dir(bdev);
    char *tmp = NULL;
    FILE *f = NULL;
    int m, n;
    char buf[1024];
    int errnum;

//    return -1;

    *len = 1024;
    memset(buf, '\0', sizeof(buf));

    if (asprintf(&tmp, "%s/device/%s", sysfs_dir, attr) < 0) 
        goto err;

    if (!(f = fopen(tmp, "r")))
        goto err;
    free(tmp);
    tmp = NULL;

    if (!fgets(buf, 1023, f))
        goto err;

    fclose(f);
    f = NULL;

//    fprintf(stderr, "%s got buf: %s\n", __func__, memdumpa(buf, strlen(buf)+1));
    for (m = 0, n = 0; buf[m]; m++) {
        if (buf[m] != ' ')
            buf[n++] = buf[m];
    }
    buf[n] = '\0';
    if (n >= 1 && buf[n-1] == '\n')
        buf[--n] = '\0';

    if (!(tmp = strdup(buf)))
        goto err;

    *value = tmp;
    *len = n;
    return 0;   
err:
    errnum = errno;
    if (tmp)
        free(tmp);
    if (f)
        fclose(f);
    errno = errnum;
    return -1;
}

static void scsi_destroy_priv(void *privp)
{
    struct scsi_probe_priv *priv = privp;

    if (priv) {
        if (priv->vendor)
            free(priv->vendor);
        if (priv->model)
            free(priv->model);
        if (priv->unique_id_80)
            free(priv->unique_id_80);
        if (priv->unique_id_83)
            free(priv->unique_id_83);
        free(priv);
    }
}

static int scsi_do_probes(struct bdevid_bdev *bdev, void **privp)
{
    struct scsi_probe_priv *priv;

    if (!privp)
        return -1;
    priv = *privp;
    if (!priv && !(priv = calloc(1, sizeof (*priv))))
        return -1;

    /* XXX it'd be nice if the kernel were reliable here, since it's got a
       white/black/quirk list, but it isn't... */
    if (!priv->vendor)
        scsi_get_devattr(bdev, "vendor", &priv->vendor, &priv->vendor_len);
    if (priv->vendor) {
        if (priv->vendor_len == 0) {
            free(priv->vendor);
            priv->vendor = NULL;
            priv->vendor_len = 0;
        }
    }

    /* XXX it'd be nice if the kernel were reliable here, since it's got a
       white/black/quirk list, but it isn't... */
    if (!priv->model)
        scsi_get_devattr(bdev, "model", &priv->model, &priv->model_len);
    if (priv->model) {
        if (priv->model_len == 0) {
            free(priv->model);
            priv->model = NULL;
            priv->model_len = 0;
        }
    }

    int fd = bdevid_bdev_get_fd(bdev);
    unsigned char page0[SCSI_INQ_BUFF_LEN];
    unsigned char inq[SCSI_INQ_BUFF_LEN];
    int m, n;
    int errnum;

    memset(inq, '\0', SCSI_INQ_BUFF_LEN);
    if ((m = scsi_inquiry(fd, 0, 0, inq, SCSI_INQ_BUFF_LEN)) < 0)
        return m;

    if (!priv->vendor) {
        if (!(priv->vendor = calloc(VENDOR_LENGTH+1, sizeof(char))))
            goto out;
        n = 0;

//        fprintf(stderr, "%s vendor: %s\n", __func__, memdumpa(inq+VENDOR_START, VENDOR_LENGTH));
        for (m = VENDOR_START; m < VENDOR_END; m++) {
            if (inq[m] != ' ')
                priv->vendor[n++] = inq[m];
        }
//        fprintf(stderr, "%s priv->vendor: %s (n=%d)\n", __func__, memdumpa(priv->vendor, VENDOR_LENGTH), n);
        if (n > 0) {
            priv->vendor[n] = '\0';
            priv->vendor_len = n;
        } else {
            free(priv->vendor);
            priv->vendor = NULL;
            priv->vendor_len = 0;
        }
    }
    if (!priv->model) {
        if (!(priv->model = calloc(MODEL_LENGTH+1, sizeof(char))))
            goto out;
        n = 0;
//        fprintf(stderr, "%s model: %s\n", __func__, memdumpa(inq+MODEL_START, MODEL_LENGTH));
        for (m = MODEL_START; m < MODEL_END; m++) {
            if (inq[m] != ' ')
                priv->model[n++] = inq[m];
        }
        if (n > 0) {
            priv->model[n] = '\0';
            priv->model_len = n;
        } else {
            free(priv->model);
            priv->model = NULL;
            priv->model_len = 0;
        }
    }

    memset(page0, '\0', SCSI_INQ_BUFF_LEN);
    if (do_scsi_page0_inquiry(fd, inq, page0, SCSI_INQ_BUFF_LEN) < 0)
        goto out;

    int ind;
    for (ind = 4; ind <= page0[3] + 3; ind++) {
        if (page0[ind] == PAGE_83) {
            do_scsi_page83_inquiry(fd, priv);
            if (result_ok(priv, 0))
                goto out;
        }
    }
    for (ind = 4; ind <= page0[3] + 3; ind++) {
        if (page0[ind] == PAGE_80) {
            do_scsi_page80_inquiry(fd, priv);
            if (result_ok(priv, 0))
                goto out;
        }
    }

out:

    errnum = errno;
    errno = 0;
    if (priv && result_ok(priv, 1)) {
        /* now we have to check and see if the kernel jammed the vendor
         * and model into the unique id for absolutely no reason at all.
         * arrrrgh.
         */
        if (priv->unique_id_83) {
            if (priv->vendor && priv->vendor[0]) {
                int l = strlen(priv->vendor);
                if (!strncmp(priv->vendor, priv->unique_id_83, l)) {
                    if (priv->unique_id_83[l] == '-')
                        l++;
                    memmove(priv->unique_id_83, priv->unique_id_83 + l,
                            strlen(priv->unique_id_83) - l + 1);
                }
            }
            if (priv->model && priv->model[0]) {
                int l = strlen(priv->model);
                if (!strncmp(priv->model, priv->unique_id_83, l)) {
                    if (priv->unique_id_83[l] == '-')
                        l++;
                    memmove(priv->unique_id_83, priv->unique_id_83 + l,
                            strlen(priv->unique_id_83) - l + 1);
                }
            }
        }
        *privp = priv;
        return 0;
    }
    scsi_destroy_priv(priv);
    errno = errnum;
    return -1;
}

static int scsi_get_vendor(struct bdevid_bdev *bdev, char **vendor,
        void **priv)
{
    if (scsi_do_probes(bdev, priv) < 0)
        return -1;
//    fprintf(stderr, "%s: *priv is %p\n", __func__, *priv);

    *vendor = (*(struct scsi_probe_priv **)priv)->vendor;
    return 0;
}

static int scsi_get_model(struct bdevid_bdev *bdev, char **model, void **priv)
{
    if (scsi_do_probes(bdev, priv) < 0)
        return -1;
//    fprintf(stderr, "%s: *priv is %p\n", __func__, *priv);

    *model = (*(struct scsi_probe_priv **)priv)->model;
    return 0;
}
static int scsi_get_unique_id(struct bdevid_bdev *bdev, char **id,
        void **privp)
{
    struct scsi_probe_priv *priv;
    if (scsi_do_probes(bdev, privp) < 0)
        return -1;
//    fprintf(stderr, "%s: *priv is %p\n", __func__, *privp);

    priv = *privp;
    if (priv) {
        if (priv->unique_id_83)
            *id = priv->unique_id_83;
        else if (priv->unique_id_80)
            *id = priv->unique_id_80;
        if (*id)
            return 0;
    }
    return -1;
}

static struct bdevid_probe_ops scsi_probe_ops = {
    .name = "scsi_probe",
    .get_vendor = scsi_get_vendor,
    .get_model = scsi_get_model,
    .get_unique_id = scsi_get_unique_id,
    .destroy_priv = scsi_destroy_priv,
};

static int scsi_init(struct bdevid_module *bm,
    bdevid_register_func register_probe)
{
    if (register_probe(bm, &scsi_probe_ops) == -1)
        return -1;
    return 0;
}

static struct bdevid_module_ops scsi_module = {
    .magic = BDEVID_MAGIC,
    .name = "scsi",
    .init = scsi_init,
};

BDEVID_MODULE(scsi_module);

/*
 * vim:ts=8:sw=4:sts=4:et
 */
