/*
 * sysfs.c - utility functions for finding things from sysfs
 *
 * Peter Jones (pjones@redhat.com)
 *
 * Copyright 2006 Red Hat, Inc.
 *
 * This software may be freely redistributed under the terms of the GNU
 * General Public License, version 2.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#define _GNU_SOURCE 1

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <dirent.h>
#include <errno.h>
#include <unistd.h>

#include <nash.h>

#include "bdevid.h"
#include "util.h"

struct bdevid_sysfs_node {
    char *dir;
};

void bdevid_sysfs_free_node(struct bdevid_sysfs_node *node)
{
    if (node) {
        if (node->dir) {
            free(node->dir);
            node->dir = NULL;
        }
        free(node);
    }
}

static int
parse_sysfs_devno(const char *path, dev_t *dev)
{
    char *first = NULL, *second;
    int major, minor;
    int fd, len = strlen(path);
    char devname[strlen(path) + 5];

    strcpy(devname, path);
    strcat(devname, "/dev");
    fd = open(devname, O_RDONLY);
    if (fd < 0) {
        if (errno == ENOENT)
            return 0;
        return fd;
    }

    len = readFD(fd, &first);

    close(fd);

    second = strchr(first, ':');
    *second++ = '\0';

    errno = 0;
    major = strtol(first, NULL, 10);

    errno = 0;
    minor = strtol(second, NULL, 10);
    free(first);

    *dev = makedev(major, minor);
    return 1;
}

static char *find_device_by_devno(char *path, dev_t goal)
{
    int ret;
    struct stat sb;
    dev_t devno;
    char *answer = NULL;

    stat(path, &sb);
    if (!S_ISDIR(sb.st_mode))
        return 0;

    ret = parse_sysfs_devno(path, &devno);
    if (ret && devno == goal)
        return strdup(path);

    if (ret == 0) {
        DIR *dir;
        struct dirent *dent;
        char *subdir;

        if (!(dir = opendir(path)))
            return NULL;

        while ((dent = readdir(dir))) {
            if (!strcmp(dent->d_name, ".") || !strcmp(dent->d_name, ".."))
                continue;

            if (asprintf(&subdir, "%s/%s", path, dent->d_name) < 0)
                break;

            answer = find_device_by_devno(subdir, goal);

            free(subdir);
            if (answer)
                break;
        }
        closedir(dir);
    }
    return answer;
}

struct bdevid_sysfs_node *bdevid_sysfs_find_node(char *file)
{
    struct stat sb;
    char *sysfs_dir;
    struct bdevid_sysfs_node *node;

    if (stat(file, &sb) < 0)
        return NULL;

    if (!(sysfs_dir = find_device_by_devno("/sys/block", sb.st_rdev)))
        return NULL;

    if (!(node = calloc(1, sizeof(*node)))) {
        free(sysfs_dir);
        return NULL;
    }

    node->dir = strdup(sysfs_dir + strlen("/sys/block"));
    free(sysfs_dir);
    return node;
}

char *bdevid_sysfs_get_name(struct bdevid_sysfs_node *node)
{
    /* XXX wrong wrong wrong */
    return node->dir;
}
char *bdevid_sysfs_get_dir(struct bdevid_sysfs_node *node)
{
    /* XXX sucky sucky */
    return node->dir;
}

/*
 * vim:ts=8:sw=4:sts=4:et
 */
