/*
 * dict.c -- simple dictionaries
 *
 * Peter Jones (pjones@redhat.com)
 *
 * Copyright 2007 Red Hat, Inc.
 *
 * This software may be freely redistributed under the terms of the GNU
 * General Public License, version 2.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define _GNU_SOURCE 1

#include "list.h"
#include "util.h"

#include <stdint.h>
#include <stdlib.h>
#include <string.h>

struct nash_dict_item {
    void *key;
    void *value;
    struct nash_dict_item *next;
};

struct nash_dict {
    nash_hash_item_func hash;
    nash_cmp_item_func cmp;
    struct nash_list *keys;
    struct nash_list *items;
};

void nash_dict_free(struct nash_dict *dict)
{
    struct nash_dict_item *item;
    if (dict) {
        if (dict->items) {
            while ((item = nash_list_get(dict->items, 0))) {
                while (item) {
                    struct nash_dict_item *next = item->next;
                    free(item);
                    item = next;
                }
                nash_list_delete(dict->items, 0);
            }
            nash_list_destroy(dict->items);
            dict->items = NULL;
        }
        if (dict->keys) {
            nash_list_destroy(dict->keys);
            dict->keys = NULL;
        }
        free(dict);
    }
}

struct nash_dict *nash_dict_new(nash_hash_item_func hash,
                                nash_cmp_item_func cmp)
{
    struct nash_dict *dict = NULL;
    int errnum;

    if (!(dict = calloc(1, sizeof (*dict))))
        return NULL;

    if (!(dict->keys = nash_dkey_list_new())) {
        save_errno(errnum, nash_dict_free(dict));
        return NULL;
    }

    if (!(dict->items = nash_void_list_new())) {
        save_errno(errnum, nash_dict_free(dict));
        return NULL;
    }

    dict->hash = hash;
    dict->cmp = cmp;
    return dict;
}

static void nash_dict_del_ikey(struct nash_dict *dict, void *keyhash)
{
    struct nash_dkey_list_item skey;
    int kpos;

    skey.key = keyhash;
    if ((kpos = nash_list_search(dict->keys, &skey)) < 0)
        return;

    nash_list_delete(dict->keys, kpos);
}

static int nash_dict_get_ikey(struct nash_dict *dict, void *keyhash,
        int create)
{
    struct nash_dkey_list_item skey, *kkey;
    int kpos;

    skey.key = keyhash;

    if ((kpos = nash_list_search(dict->keys, &skey)) < 0) {
        if (!create)
            return kpos;

        if (!(kkey = calloc(1, sizeof (*kkey))))
            return kpos;

        kkey->key = keyhash;
        kkey->position = nash_list_len(dict->items);
        if ((kpos = nash_list_add_ref(dict->keys, kkey)) < 0) {
            int errnum = errno;
            free(kkey);
            errno = errnum;
            return kpos;
        }
    } else {
        kkey = nash_list_get(dict->keys, kpos);
    }

    return kkey->position;
}

int nash_dict_get_kv(struct nash_dict *dict, void *key, void **k, void **v)
{
    int ikey;
    struct nash_dict_item *item;
    void *keyhash;

    keyhash = (void *)(unsigned long)dict->hash(key);
    if ((ikey = nash_dict_get_ikey(dict, keyhash, 0)) < 0)
        return -1;

    item = nash_list_get(dict->items, ikey);
    while (item) {
        if (!dict->cmp(&item->key, &key)) {
            *k = item->key;
            *v = item->value;
            return 0;
        }

        item = item->next;
    }
    return -1;
}

void *nash_dict_get(struct nash_dict *dict, void *key)
{
    void *k, *v = NULL;;

    nash_dict_get_kv(dict, key, &k, &v);
    return v;
}

int nash_dict_add(struct nash_dict *dict, void *key, void *value)
{
    int ikey, rc, errnum;
    struct nash_dict_item *item, *newitem;
    void *keyhash;

    keyhash = (void *)(unsigned long)dict->hash(key);
    if ((ikey = nash_dict_get_ikey(dict, keyhash, 1)) < 0)
        return ikey;

    newitem = item = nash_list_get(dict->items, ikey);
    while (item) {
        if (!dict->cmp(&item->key, &key)) {
            item->key = key;
            item->value = value;
            return 0;
        }
        item = item->next;
    }
    item = newitem;

    if (!(newitem = calloc(1, sizeof (*newitem))))
        return -1;
    newitem->key = key;
    newitem->value = value;
    /* if there was already an item in this position, just add the new
     * one into the chain.  These should be short enough that order isn't
     * crucial to find it later. */
    if (item) {
        newitem->next = item->next;
        item->next = newitem;
        return 0;
    }
    /* this leaves *keys having one extra entry, but it doesn't matter
     * since nash_dict_get_ikey() gives the length of ->items, so we'll
     * re-use it anyway */
    if ((rc = nash_list_add_ref(dict->items, newitem)) < 0)
        save_errno(errnum, free(newitem));

    return rc;
}

/* XXX this is just a bit too trusting, but... */
int nash_dict_update(struct nash_dict *dict, uint32_t oldhash, uint32_t newhash,
        void *key, void *value)
{
    int oldikey, newikey;
    struct nash_dict_item *newitem, *item, *next, *prev;
    int ret, errnum;

    if ((oldikey = nash_dict_get_ikey(dict, (void *)(unsigned long)oldhash, 0)) < 0)
        return nash_dict_add(dict, key, value);
    if ((newikey = nash_dict_get_ikey(dict, (void *)(unsigned long)newhash, 1)) < 0)
        return -1;

    if (!(item = nash_list_get(dict->items, oldikey)))
        return -1;

    next = prev = NULL;
    while (item) {
        next = item->next;

        if (item->key == key)
            break;

        prev = item;
        item = next;
        next = NULL;
    }
    if (!item)
        return -1;

    /* now unstitch it ... */
    /* If we're last, next will be null.  If we're first, prev will be. */
    if (!prev) {
        if ((ret = nash_list_replace(dict->items, oldikey, next)) < 0)
            return -1;
    } else {
        prev->next = next;
    }
    item->key = key;
    item->value = value;
    item->next = NULL;
    newitem = item;

    /* ... and put it back in */
    item = nash_list_get(dict->items, newikey);
    if (item) {
        newitem->next = item->next;
        item->next = newitem;
        return 0;
    }

    if (!(ret = nash_list_add_ref(dict->items, item)))
        save_errno(errnum, free(item));
    return ret;
}

int nash_dict_del(struct nash_dict *dict, void *key)
{
    int ikey;
    struct nash_dict_item *item, *prev, *next;
    void *keyhash;

    keyhash = (void *)(unsigned long)dict->hash(key);
    if ((ikey = nash_dict_get_ikey(dict, keyhash, 0)) < 0)
        return ikey;

    if (!(item = nash_list_get(dict->items, ikey)))
        return -1;

    next = prev = NULL;
    while (item) {
        next = item->next;

        if (!dict->cmp(&item->key, &key))
            break;
        
        prev = item;
        item = next;
        next = NULL;
    }
    if (!item)
        return -1;

    free(item);

    /* If we're last, next will be null.  If we're first, prev will be. */
    if (!prev) {
        /* If it's a one-item list, "next" is NULL */
        if (!next)
            nash_dict_del_ikey(dict, keyhash);
        return nash_list_replace(dict->items, ikey, next);
    }

    prev->next = next;
    return 0;
}

struct nash_dict_iter {
    int ikey;
    int inum;
};

void nash_dict_items_end(struct nash_dict_iter *iter)
{
    memset(iter, '\0', sizeof (*iter));
    free(iter);
}

struct nash_dict_iter *nash_dict_items_next(struct nash_dict *dict,
        struct nash_dict_iter *iter, void **key, void **value)
{
    struct nash_dict_item *item;
    int nkeys = nash_list_len(dict->items);

    if (!iter) {
        if (!nkeys)
            return NULL;
        if (!(iter = calloc(1, sizeof(*iter))))
            return iter;
        iter->ikey = 0;
        iter->inum = 0;
    }
    while (iter->ikey < nkeys) {
        int j = ++iter->inum;
        if (!(item = nash_list_get(dict->items, iter->ikey)))
            continue;
        while (--j) {
            if (!(item = item->next)) {
                iter->ikey++;
                iter->inum = 0;
                break;
            }
        }
        if (iter->inum > 0) {
            *key = item->key;
            *value = item->value;
            return iter;
        }
    }
    free(iter);
    return NULL;
}

/*
 * vim:ts=8:sw=4:sts=4:et
 */
