/*
 * procdev.c -- functions to deal with /proc/devices and related files.
 *
 * Peter Jones (pjones@redhat.com)
 *
 * Copyright 2007 Red Hat, Inc.
 *
 * This software may be freely redistributed under the terms of the GNU
 * General Public License, version 2.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define _GNU_SOURCE 1
#include "lib.h"

static struct proc_dev_info proc_dev_info[] = {
    /* these names needs to match the list in devtree.c */
    {   "ataraid",
        S_IFBLK,
        DEV_TYPE_DISK,
        DEV_NAME_BEGINS|DEV_MUST_HAVE_PARTS|DEV_USES_P_CONV, /* XXX ? */
        "ataraid/",
    },
    {   "cciss",
        S_IFBLK,
        DEV_TYPE_DISK,
        DEV_NAME_BEGINS|DEV_MUST_HAVE_PARTS|DEV_USES_P_CONV,
        "cciss/",
    },
    {   "dac960",
        S_IFBLK,
        DEV_TYPE_DISK,
        DEV_NAME_EQUALS|DEV_MUST_HAVE_PARTS|DEV_USES_P_CONV,
        "rd/",
    },
    {   "device-mapper",
        S_IFCHR,
        DEV_TYPE_NONE,
        DEV_NAME_EQUALS,
        "",
    },
    {   "device-mapper",
        S_IFBLK,
        DEV_TYPE_DM,
        DEV_NAME_EQUALS|DEV_MUST_HAVE_PARTS|DEV_USES_P_CONV,
        "mapper/"
    },
    {   "floppy",
        S_IFBLK,
        DEV_TYPE_FLOPPY,
        DEV_NAME_EQUALS,
        "",
    },
    {   "i2o_block",
        S_IFBLK,
        DEV_TYPE_DISK,
        DEV_NAME_EQUALS|DEV_MUST_HAVE_PARTS,
        "i2o/",
    },
    {   "ida",
        S_IFBLK,
        DEV_TYPE_DISK,
        DEV_NAME_BEGINS|DEV_MUST_HAVE_PARTS|DEV_USES_P_CONV,
        "ida/",
    },
    {   "ide",
        S_IFBLK,
        DEV_TYPE_DISK,
        DEV_NAME_BEGINS|DEV_MUST_HAVE_PARTS,
        "",
    },
    {   "iseries/vd",
        S_IFBLK,
        DEV_TYPE_DISK,
        DEV_NAME_EQUALS|DEV_MUST_HAVE_PARTS,
        "iseries/",
    },
#if 0 || defined(NASH_SUPPORT_ABSURD_DEVICE_TOPOLOGY)
    {   "md",
        S_IFBLK,
        DEV_TYPE_MD_RAID,
        DEV_NAME_EQUALS|DEV_MAY_HAVE_PARTS|DEV_USES_P_CONV,
        "",
    {   "mdp",
        S_IFBLK,
        DEV_TYPE_PARTITION,
        DEV_NAME_EQUALS,
        "",
    },
#else
    {   "md",
        S_IFBLK,
        DEV_TYPE_MD_RAID,
        DEV_NAME_EQUALS,
        ""
    },
#endif
    {   "sd",
        S_IFBLK,
        DEV_TYPE_DISK,
        DEV_NAME_EQUALS|DEV_MUST_HAVE_PARTS,
        "",
    },
    {   "sx8",
        S_IFBLK,
        DEV_TYPE_DISK,
        DEV_NAME_BEGINS|DEV_MUST_HAVE_PARTS|DEV_USES_P_CONV,
        "sx8/",
    },
    { "", 0, 0, 0, "" }
};

struct dev_type {
    char name[64];
    int mode;
    long major;
    long minor;
};

static uint32_t dev_type_name_hash(const void *dtp)
{
    struct dev_type *dt = (struct dev_type *)dtp;
    uint32_t adler = 0;

    adler = adler32_int(adler, dt->mode);
    adler = adler32_str(adler, dt->name);
    return adler;
}

static uint32_t dev_type_devno_hash(const void *dtp)
{
    struct dev_type *dt = (struct dev_type *)dtp;
    uint32_t adler = 0;

    adler = adler32_int(adler, dt->major);
    adler = adler32_int(adler, dt->minor);
    return adler;
}

static int dev_type_dict_cmp(const void *a, const void *b)
{
    struct dev_type *d0 = *(struct dev_type **)a;
    struct dev_type *d1 = *(struct dev_type **)b;
    int rc;

    if (!d0 || !d1)
        return d0 ? -1 : (d1 ? 1 : -1 );

    if (d0->mode != d1->mode)
        return d0->mode - d1->mode;

    if ((rc = strcmp(d0->name, d1->name)))
        return rc;

    if (d0->major != d1->major)
        return d0->major - d1->major;

    if (d0->minor != d1->minor)
        return d0->minor - d1->minor;

    return 0;
}

struct nash_proc_dev_cache {
    uint32_t proc_devices_hash;
    struct nash_dict *char_devs;
    struct nash_dict *block_devs;

    uint32_t proc_misc_hash;
    struct nash_dict *proc_misc;

    struct nash_dict *devnames;
    struct nash_dict *devnos;
};

struct proc_file_item {
    char name[64];
    struct nash_list *numbers;
};

static void nash_pdc_empty_cache(struct nash_dict **dp)
{
    struct nash_dict *d = *dp;
    struct nash_dict_iter *iter = NULL;
    void *k, *v;

    if (!d)
        return;

    while ((iter = nash_dict_items_next(d, iter, &k, &v))) {
        struct proc_file_item *pfi = (struct proc_file_item *)v;

        nash_list_destroy(pfi->numbers);
        free(pfi);
    }
    nash_dict_free(d);
}

void nash_proc_dev_cache_destroy(nashContext *nc)
{
    if (nc->pdc) {
        struct nash_proc_dev_cache *pdc = nc->pdc;
        
        if (pdc->devnames) {
            struct nash_dict_iter *iter = NULL;
            struct dev_type *d;
            void *k;
            while ((iter = nash_dict_items_next(pdc->devnames, iter, &k, (void *)&d)))
                free(d);
            nash_dict_free(pdc->devnames);
            pdc->devnames = NULL;
        }
        nash_dict_free(pdc->devnos);
        pdc->devnames = NULL;
        nash_pdc_empty_cache(&pdc->block_devs);
        nash_pdc_empty_cache(&pdc->char_devs);
        nash_pdc_empty_cache(&pdc->proc_misc);
        free(nc->pdc);
        nc->pdc = NULL;
    }
}

static void add_pdc_dev(struct nash_dict *devnames, struct nash_dict *devnos,
        const char *name, int mode, long major, long minor)
{
    struct dev_type k, *dt;

    memset(&k, '\0', sizeof (k));
    k.mode = mode;
    strncpy(k.name, name, sizeof(k.name));
    k.major = major;
    k.minor = minor;

    if ((dt = nash_dict_get(devnames, &k)))
        return;

    if (!(dt = calloc(1, sizeof (*dt))))
        return;

    memmove(dt, &k, sizeof (*dt));

    if (nash_dict_add(devnames, dt, dt) < 0) {
        free(dt);
        return;
    }
    if (nash_dict_add(devnos, dt, dt) < 0) {
        nash_dict_del(devnames, dt);
        free(dt);
    }
}

static void add_missing_types(struct nash_dict *devnames,
        struct nash_dict *devnos)
{
    int i;
    const struct dev_type missing_dev_types[] = {
        /* pre-seeded with things that don't show up in /proc/devices :/ */
#if 0
        { "ramdisk0", S_IFBLK, 0, -1 },
        { "ramdisk", S_IFBLK, 1, -1 },
#endif
        { "floppy", S_IFBLK, 2, -1 },
        { "ide0", S_IFBLK, 3, -1 },
        { "ide1", S_IFBLK, 22, -1 },
        { "ide2", S_IFBLK, 33, -1 },
        { "ide3", S_IFBLK, 34, -1 },
        { "ide4", S_IFBLK, 56, -1 },
        { "ide5", S_IFBLK, 57, -1 },
        { "ide6", S_IFBLK, 88, -1 },
        { "ide7", S_IFBLK, 89, -1 },
        { "ide8", S_IFBLK, 90, -1 },
        { "ide9", S_IFBLK, 91, -1 },
        { "ataraid", S_IFBLK, 114, -1 },
        { "", -1, 0, 0 }
    };

    for (i = 0; missing_dev_types[i].mode != -1; i++) {
        struct dev_type const *d = &missing_dev_types[i];

        add_pdc_dev(devnames, devnos, d->name, d->mode, d->major, d->minor);
    }
}

int nash_proc_dev_cache_init(nashContext *nc)
{
    int errnum;

    if (!nc->pdc) {
        if (!(nc->pdc = calloc(1, sizeof (*nc->pdc))))
            return -1;
        
        if (!(nc->pdc->char_devs = nash_dict_new(stringhash, stringsort))) {
err:
            save_errno(errnum, nash_proc_dev_cache_destroy(nc));
            return -1;
        }

        if (!(nc->pdc->block_devs = nash_dict_new(stringhash, stringsort)))
            goto err;

        if (!(nc->pdc->proc_misc = nash_dict_new(stringhash, stringsort)))
            goto err;

        if (!(nc->pdc->devnames = nash_dict_new(dev_type_name_hash,
                        dev_type_dict_cmp)))
            goto err;
        if (!(nc->pdc->devnos = nash_dict_new(dev_type_devno_hash,
                        dev_type_dict_cmp)))
            goto err;

        nc->pdc->proc_devices_hash = 0;
        nc->pdc->proc_misc_hash = 0;
    }

    return 0;
}

static uint32_t pfihash(void *pfip)
{
    struct proc_file_item *pfi = (struct proc_file_item *)pfip;

    return stringhash(pfi->name);
}

static uint32_t pfisort(void *pfip0, void *pfip1)
{
    struct proc_file_item *pfi0 = (struct proc_file_item *)pfip0;
    struct proc_file_item *pfi1 = (struct proc_file_item *)pfip1;

    return strcmp(pfi0->name, pfi1->name);
}

static void test_should_update(char *path, char **buf, uint32_t *old_hash)
{
    uint32_t new_hash;
    int fd;
    int rc;

    *buf = NULL;
    if ((fd = open(path, O_RDONLY)) < 0)
        return;

    rc = readFD(fd, buf);
    close(fd);
    if (rc < 0)
        return;

    new_hash = adler32_str(0, *buf);
    if (*old_hash == new_hash) {
        free(*buf);
        *buf = NULL;
        return;
    }

    *old_hash = new_hash;
    return;
}

typedef enum {
        CHAR,
        BLOCK,
        MISC,
} state_t;

static int get_state(char *buf) {
    switch (buf[0]) {
        case 'C':
            return CHAR;
        case 'B':
            return BLOCK;
    }
    return MISC;
}

static void add_dev_info(struct nash_proc_dev_cache *pdc, int state,
        int devno, char *name)
{
    struct nash_dict *dict;
    struct proc_file_item *item;

    switch (state) {
        case BLOCK:
            dict = pdc->block_devs;
            break;
        case CHAR:
            dict = pdc->char_devs;
            break;
        case MISC:
            dict = pdc->proc_misc;
            break;
    };

    if (!(item = nash_dict_get(dict, name))) {
        if (!(item = calloc(1, sizeof (*item))))
            return;
        strcpy(item->name, name);
        if (!(item->numbers = nash_ptr_list_new())) {
            free(item);
            return;
        }
        nash_list_add(item->numbers, (void *)(unsigned long)devno);
        nash_dict_add(dict, name, item);
    } else {
        nash_list_add(item->numbers, (void *)(unsigned long)devno);
    }
}

static void update_pdc_devs(struct nash_proc_dev_cache *pdc)
{
    struct proc_file_item *item;
    long major;
    struct nash_dict_iter *iter = NULL;
    void *name;
    int i = 0;

    while ((iter = nash_dict_items_next(pdc->block_devs, iter, &name,
                    (void **)&item))) {
        i = 0;
        while ((major = (long)nash_list_get(item->numbers, i++))) {
            add_pdc_dev(pdc->devnames, pdc->devnos, item->name, S_IFBLK,
                    major, -1);
        }
    }

    while ((iter = nash_dict_items_next(pdc->char_devs, iter, &name,
                    (void **)&item))) {
        i = 0;
        while ((major = (long)nash_list_get(item->numbers, i++))) {
            if (!strcmp(item->name, "misc")) {
                struct nash_dict_iter *misc_iter = NULL;
                long misc_major = major;

                while ((misc_iter = nash_dict_items_next(pdc->proc_misc,
                                misc_iter, &name, (void **)&item))) {
                    i = 0;
                    while ((major = (long)nash_list_get(item->numbers, i++))) {
                        add_pdc_dev(pdc->devnames, pdc->devnos, item->name,
                                S_IFCHR, misc_major, major);
                    }
                }
            } else {
                i = 0;
                while ((major = (long)nash_list_get(item->numbers, i++))) {
                    add_pdc_dev(pdc->devnames, pdc->devnos, item->name,
                            S_IFCHR, major, -1);
                }
            }
        }
    }
}

void nash_proc_dev_cache_update(nashContext *nc)
{
    char *pd_buf = NULL;
    char *pm_buf = NULL;
    char *bufs[3];
    int i = 0;
    struct nash_proc_dev_cache *pdc;
    state_t state, newstate;

    if (nash_proc_dev_cache_init(nc) < 0)
        return;
    pdc = nc->pdc;
//    pdc->proc_devices_hash = 0;

    test_should_update("/proc/devices", &pd_buf, &pdc->proc_devices_hash);
    test_should_update("/proc/misc", &pm_buf, &pdc->proc_misc_hash);


    if (pd_buf)
        bufs[i++] = pd_buf;
    if (pm_buf)
        bufs[i++] = pm_buf;
    bufs[i] = NULL;

    if (!pd_buf && !pm_buf)
        return;

    if (pdc->devnames) {
        struct nash_dict_iter *iter = NULL;
        struct dev_type *d;
        void *k;
        while ((iter = nash_dict_items_next(pdc->devnames, iter, &k, (void *)&d)))
                free(d);
        nash_dict_free(pdc->devnames);
        pdc->devnames = NULL;
    }
    nash_dict_free(pdc->devnos);
    pdc->devnos = NULL;
    if (pd_buf) {
        nash_pdc_empty_cache(&pdc->block_devs);
        pdc->block_devs = nash_dict_new(stringhash, stringsort);
        nash_pdc_empty_cache(&pdc->char_devs);
        pdc->char_devs = nash_dict_new(stringhash, stringsort);
    }
    if (pm_buf) {
        nash_pdc_empty_cache(&pdc->proc_misc);
        pdc->proc_misc = nash_dict_new(stringhash, stringsort);
    }
    if (!(pdc->devnames = nash_dict_new(dev_type_name_hash, dev_type_dict_cmp))
            || !(pdc->devnos = nash_dict_new(dev_type_devno_hash, 
                    dev_type_dict_cmp))) {
        nash_dict_free(pdc->devnames);
        pdc->devnames = NULL;
        nash_dict_free(pdc->devnos);
        pdc->devnos = NULL;
    }
    if (pdc->devnames) {
        add_missing_types(pdc->devnames, pdc->devnos);
        for (i = 0; bufs[i]; i++) {
            char *buf = bufs[i];
            size_t o = 0;
    
            state = get_state(buf);
    
            while (buf[o]) {
                char name[64] = {0};
                long devno;
    
                while (buf[o] == '\n')
                    o++;
    
                newstate = get_state(buf+o);
                if (newstate != MISC) {
                    state = newstate;
                } else {
                    if (sscanf(buf+o, "%ld %s\n", &devno, name) < 2)
                        continue;
                    add_dev_info(pdc, state, devno, name);
                }
                o += strcspn(buf+o, "\n");
            }
        }
        update_pdc_devs(pdc);
    }
    if (pd_buf)
        free(pd_buf);
    if (pm_buf)
        free(pm_buf);
}

static int _dev_name_cmp(struct dev_type *dev_type,
        char *name, char **newname, struct proc_dev_info const **info)
{
    const struct proc_dev_info *i;
    int l, m, n;
    int rc;
    
    l = strlen(name);
    m = strlen(dev_type->name);
    /* exit early if the largest common subset doesn't match */
    if ((rc = strncmp(name, dev_type->name, MIN(l,m))))
        return rc;
    /* steps:
     * 1) find the proc_dev_info entry that corresponds with dev_type
     * 2) if it exists, see if that matches the name
     *    else, see if dev_type->name matches exactly
     * 3) set new if appropriate
     * 4) return comparison
     */
    for (i = &proc_dev_info[0]; i->name[0] != '\0'; i++) {
        if (i->mode != dev_type->mode)
            continue;

        n = strlen(i->name);
        if (!(rc = strncmp(dev_type->name, i->name,
                    ((i->flags & DEV_NAME_BEGINS) ? n : MAX(m,n))))) {
            /* we have a match */
            if (!rc) {
                *info = i;
                *newname = (char *)&i->name[0];
            }
            return rc;
        }
    }
    rc = strcmp(dev_type->name, name);
    if (!rc) {
        *info = NULL;
        *newname = dev_type->name;
    }
    return rc;
}

int getDevsFromProc(nashContext *nc, int pos, int mode, char **name,
        dev_t *devno, struct proc_dev_info const **info)
{
    struct nash_dict *dict;
    struct nash_dict_iter *iter;
    int i;
    struct dev_type *d;
    void *v;
    struct proc_dev_info const *pdi;
    char *newname;
    errno = 0;

    if (pos < -1 || pos > 1022)
        return -1;
    if ((!name || !*name) && !devno)
        return -1;

    if (pos == -1)
        nash_proc_dev_cache_update(nc);

    if (!nc->pdc)
        return -1;

    dict = name && *name ? nc->pdc->devnames : nc->pdc->devnos;
    i = -1;
    iter = NULL;
    for (i = -1; i < pos; i++) {
        if (!(iter = nash_dict_items_next(dict, iter, (void **)&d, &v)))
            break;
    }
    if (!iter && pos >= 0) 
        return -1;
    pos++;
    while ((iter = nash_dict_items_next(dict, iter, (void **)&d, &v))) {
        if (mode != -1 && mode != d->mode)
            continue;
        newname = NULL;
        pdi = NULL;
        if (*name) {
            if (!_dev_name_cmp(d, *name, &newname, &pdi)) {
                *info = pdi;
                *devno = makedev(d->major, d->minor);
                if (newname && *name != newname)
                    *name = newname;
                nash_dict_items_end(iter);
                return pos;
            }
        } else {
            long dmaj = major(*devno);
            long dmin = minor(*devno);
            if ((dmaj == -1 || d->major == -1 || dmaj == d->major) &&
                    (dmin == -1 || d->minor == -1 || dmin == d->minor)) {
                *name = d->name;
                if (!_dev_name_cmp(d, *name, &newname, &pdi)) {
                    *info = pdi;
                    *name = d->name;
                }
                nash_dict_items_end(iter);
                return pos;
            }
        }
    }
    return -1;
}

/*
 * vim:ts=8:sw=4:sts=4:et
 */
