/*
 * util.c -- a small collection of utility functions for nash and libnash.a
 *
 * Peter Jones (pjones@redhat.com)
 *
 * Copyright 2006-2007 Red Hat, Inc.
 *
 * This software may be freely redistributed under the terms of the GNU
 * General Public License, version 2.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef NASH_PRIV_UTIL_H
#define NASH_PRIV_UTIL_H 1

#ifdef _GNU_SOURCE
#define _GNU_SOURCE_DEFINED
#else
#define _GNU_SOURCE 1
#endif

#include <sys/types.h>
#include <sys/time.h>
#include <sys/param.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <dirent.h>
#include <stdio.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>
#include <stdint.h>
#include <signal.h>

#include "adler32.h"

struct nashContext_s;

extern struct nashContext_s *_nash_context;

#define nassert(x) ({ if (!(x)) raise(SIGABRT); })

#define strlimitcmp(a, b) strncmp((a), (b), strlen(a))

static inline int
setFdCoe(int fd, int enable)
{
    int rc;
    long flags = 0;

    rc = fcntl(fd, F_GETFD, &flags);
    if (rc < 0)
        return rc;

    if (enable)
        flags |= FD_CLOEXEC;
    else
        flags &= ~FD_CLOEXEC;

    rc = fcntl(fd, F_SETFD, flags);
    return rc;
}

static inline int
readFD (int fd, char **buf)
{
    char *p;
    size_t size = 16384;
    int s = 0, filesize = 0;

    if (!(*buf = calloc (16384, sizeof (char))))
        return -1;

    do {
        p = *buf + filesize;
        s = read(fd, p, 16384 - s);
        if (s < 0)
            break;
        filesize += s;
        /* only exit for empty reads */
        if (s == 0)
            break;
        else if (s == 16384) {
            *buf = realloc(*buf, size + 16384);
            memset(*buf + size, '\0', 16384);
            size += s;
            s = 0;
        } else {
            size += s;
        }
    } while (1);

    *buf = realloc(*buf, filesize+1);
    (*buf)[filesize] = '\0';

    return *buf ? filesize : -1;
}

static inline int
readFDbuf(int fd, char *buf, size_t size)
{
    char *p;
    int s = 0, filesize = 0;

    if (!buf)
        return -1;

    do {
        p = buf + filesize;
        s = read(fd, p, size - s);
        if (s < 0)
            break;
        filesize += s;
        /* only exit for empty reads or full buffer */
        if (s == 0)
            break;
    } while (filesize < size);
    return filesize;
}

static inline uint32_t
checksumFd(int fd)
{
    char buf[16384] = { 0 };
    size_t s;
    uint32_t a = 0;

    if (lseek(fd, 0, SEEK_SET) < 0) {
        return 0;
    }

    while ((s = read(fd, buf, 16384)) >= 0) {
        if (s == 0)
            return a;

        a = adler32(a, buf, s);
    }
    return a;
}

static inline uint32_t
checksumFile(char *path)
{
    int fd;
    uint32_t adler = 0;

    if ((fd = open(path, O_RDONLY)) < 0)
        return adler;

    adler = checksumFd(fd);

    close(fd);
    return adler;
}

static int
mkdirchain_helper(char *pathname, size_t n, mode_t mode, int toplevel)
{
    for (; n > 0; n--) {
        int ret = 0;
        if (pathname[n] == '/') {
            pathname[n] = '\0';
            if (access(pathname, R_OK))
                ret = mkdirchain_helper(pathname, n, mode, 0);
            pathname[n] = '/';
            if (ret == 0 && !toplevel)
                ret = mkdir(pathname, mode);
            if (ret < 0)
                return ret;
        }
    }
    return 0;
}

static int mkdirchain(char *pathname, size_t n, mode_t mode)
{
    return mkdirchain_helper(pathname, n, mode, 1);
}

static int
smartmknod(const char * device, mode_t mode, dev_t dev)
{
    char buf[PATH_MAX];
    ssize_t l, n;

    nassert(device && device[0]);

    l = MAX(MIN(strlen(device), PATH_MAX), 0);
    for (n = l; n >= 0; n--)
	buf[n] = device[n] == '!' ? '/' : device[n];

    if (!access(buf, F_OK))
        return 0;
    if (mkdirchain(buf, l, 0755) < 0)
        return -1;
    return mknod(buf, mode, dev);
}

static inline int stringsort(const void *v0, const void *v1)
{
    const char * const *s0=v0, * const *s1=v1;
    return strcoll(*s0, *s1);
}

static inline uint32_t stringhash(const void *s)
{
    return adler32_str(0, s);
}

#define USECS_PER_SEC 1000000LL
#define NSECS_PER_USEC 1000LL
#define NSECS_PER_SEC (NSECS_PER_USEC * USECS_PER_SEC)

static inline void
nsectospec(long long nsecs, struct timespec *ts)
{
    if (nsecs < 0) {
        /* If they really want longer than LONG_MAX seconds, 
         * they win the best uptime award before even running init.
         */
        ts->tv_sec = LONG_MAX;
        ts->tv_nsec = 999999999;
        return;
    }
    ts->tv_sec = nsecs / NSECS_PER_SEC;
    ts->tv_nsec = (nsecs % NSECS_PER_SEC);
}

static inline void
usectospec(long long usecs, struct timespec *ts)
{
    if (usecs > 0 && LLONG_MAX / NSECS_PER_USEC > usecs)
        usecs *= NSECS_PER_USEC;
    
    nsectospec(usecs, ts);
}

static inline int
speczero(struct timespec *ts)
{
    return (ts && ts->tv_sec == 0 && ts->tv_nsec == 0);
}

static inline int
specinf(struct timespec *ts)
{
    return !ts ||
           (ts->tv_sec < 0 || ts->tv_nsec < 0) ||
           (ts->tv_sec == LONG_MAX && ts->tv_nsec == 999999999);
}

static inline long long
spectonsec(struct timespec *ts)
{
    long long nsecs = 0;
    if (specinf(ts))
        return -1;
    
    nsecs = ts->tv_sec * NSECS_PER_SEC;
    nsecs += ts->tv_nsec;
    return nsecs;
}

static inline long long
spectousec(struct timespec *ts)
{
    long long usecs = spectonsec(ts);

    return usecs < 0 ? usecs : usecs / NSECS_PER_USEC;
}

static inline int
gettimespecofday(struct timespec *ts)
{
    struct timeval tv = {0, 0};
    int rc;

    rc = gettimeofday(&tv, NULL);
    if (rc >= 0) {
        ts->tv_sec = tv.tv_sec;
        ts->tv_nsec = tv.tv_usec * NSECS_PER_USEC;
    }
    return rc;
}

/* minuend minus subtrahend equals difference */
static inline void
tssub(struct timespec *minuend, struct timespec *subtrahend,
      struct timespec *difference)
{
    long long m, s, d;

    m = spectonsec(minuend);
    s = spectonsec(subtrahend);

    if (s < 0) {
        d = 0;
    } else if (m < 0) {
        d = -1;
    } else {
        m -= s;
        d = m < 0 ? 0 : m;
    }

    nsectospec(d, difference);
    return;
}

static inline void
tsadd(struct timespec *augend, struct timespec *addend, struct timespec *sum)
{
    long long aug, add;

    aug = spectonsec(augend);
    add = spectonsec(addend);

//    printf("aug: %Ld add: %Ld\n", aug, add);

    if (aug < 0 || add < 0)
        nsectospec(-1, sum);
    else if (LLONG_MAX - MAX(add,aug) < MAX(add,aug))
        nsectospec(LLONG_MAX, sum);
    else
        nsectospec(aug+add, sum);
    return;
}

#define tsGT(x,y) (tscmp((x), (y)) < 0)
#define tsGE(x,y) (tscmp((x), (y)) <= 0)
#define tsET(x,y) (tscmp((x), (y)) == 0)
#define tsNE(x,y) (tscmp((x), (y)) != 0)
#define tsLE(x,y) (tscmp((x), (y)) >= 0)
#define tsLT(x,y) (tscmp((x), (y)) > 0)

static inline int
tscmp(struct timespec *a, struct timespec *b)
{
    long long m, s;
    long long rc;

    m = spectonsec(a);
    s = spectonsec(b);

    if (s < 0) {
        rc = 1;
        if (m < 0)
            rc = 0;
    } else if (m < 0) {
        rc = -1;
    } else {
        rc = MIN(MAX(s-m, -1), 1);
    }

    return rc;
}


static inline void
speccpy(struct timespec *dest, struct timespec *src, long long min_nsecs,
        long long max_nsecs)
{
    struct timespec min, max;

    if (!dest)
        return;
    nsectospec(min_nsecs, &min);
    nsectospec(max_nsecs, &max);

    if (!src)
        src = &max;
    if (tsLT(src, &min))
        src = &min;
    if (tsGT(src, &max))
        src = &max;
    memcpy(dest,src,sizeof (*dest));
}

static inline void
udelayspec(struct timespec ts)
{
    if (specinf(&ts))
        usectospec(-1, &ts);
    while (nanosleep(&ts, &ts) == -1 && errno == EINTR)
        ;
}

static inline void
udelay(long long usecs)
{
    struct timespec rem = {0,0};

    usectospec(usecs, &rem);
    udelayspec(rem);
}

extern char *readlink_malloc(const char *filename);

#define readlinka(_filename) ({             \
        char *_buf0, *_buf1 = NULL;         \
        _buf0 = readlink_malloc(_filename); \
        if (_buf0 != NULL) {                \
            _buf1 = strdupa(_buf0);         \
            free(_buf0);                    \
        }                                   \
        _buf0 = _buf1;                      \
    })

#define asprintfa(str, fmt, ...) ({                 \
        char *_tmp = NULL;                          \
        int _rc;                                    \
        _rc = asprintf((str), (fmt), __VA_ARGS__);  \
        if (_rc != -1) {                            \
            _tmp = strdupa(*(str));                 \
            if (!_tmp) {                            \
                _rc = -1;                           \
            } else {                                \
                free(*(str));                       \
                *(str) = _tmp;                      \
            }                                       \
        }                                           \
        _rc;                                        \
    })

#if 0 /* not just yet */
extern int qprintf(const char *format, ...)
    __attribute__((format(printf, 1, 2)))
    __attribute__((deprecated));
extern int eprintf(const char *format, ...)
    __attribute__((format(printf, 1, 2)))
    __attribute__((deprecated));
#else
extern int qprintf(const char *format, ...)
    __attribute__((format(printf, 1, 2)));
extern int eprintf(const char *format, ...)
    __attribute__((format(printf, 1, 2)));
#endif

#include <sys/syscall.h>
#include <sys/poll.h>
#ifdef __ia64__
static inline int nash_ppoll(struct pollfd *fds, nfds_t nfds,
        struct timespec *timeout, const sigset_t *sigmask, int nsigs)
{
    int rc;
    struct timespec start, end;

    if (timeout)
        gettimespecofday(&start);
    rc = ppoll(fds, nfds, timeout, sigmask);
    if (timeout) {
        if (rc == 0) {
            usectospec(0, timeout);
        } else {
            gettimespecofday(&end);
            tssub(&end, &start, &end);
            tssub(timeout, &end, timeout);
        }
    }
    return rc;
}
#else
#define nash_ppoll(fds, nfds, timeout, sigmask, nsigs) \
        syscall(SYS_ppoll, fds, nfds, timeout, sigmask, nsigs)
#endif

#define save_errno(tmp, unsafe_code) ({     \
        (tmp) = errno ;                     \
        unsafe_code ;                       \
        errno = (tmp) ;                     \
    })

#define xfree(x) ({if (x) { free(x); x = NULL; }})
#define movptr(x, y) ({(x) = (y); (y) = NULL; })
#define xmovptr(x, y) ({if (!(y) && (x)) movptr((x),(y));})

#ifndef _GNU_SOURCE_DEFINED
#undef _GNU_SOURCE
#else
#undef _GNU_SOURCE_DEFINED
#endif

#endif /* NASH_PRIV_UTIL_H */
/*
 * vim:ts=8:sw=4:sts=4:et
 */
