/*
 * vitals.c -- abstraction for the vital data regarding a block device.
 *
 * Peter Jones (pjones@redhat.com)
 *
 * Copyright 2006 Red Hat, Inc.
 *
 * This software may be freely redistributed under the terms of the GNU
 * public license.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define _GNU_SOURCE 1

#include <sys/types.h>
#include <string.h>
#include <stdlib.h>

#include <nash.h>
#include <bdevid.h>
#include "lib.h"
#include "devtree.h"

extern struct nash_bdev_vitals *nash_vitals_cache_search(nashContext *nc,
        struct nash_bdev_vitals *vitals);
extern struct nash_bdev_vitals *nash_vitals_cache_insert(nashContext *nc,
        struct nash_bdev_vitals *vitals);
extern void nash_vitals_cache_remove(nashContext *nc,
        struct nash_bdev_vitals *vitals);

static uint32_t nash_bdev_vitals_hash(const void *vitalsp)
{
    const struct nash_bdev_vitals *vitals = vitalsp;
    uint32_t adler = 0;

    adler = adler32_int(adler, vitals->type);

    switch(vitals->type) {
        case DEV_TYPE_DISK:
            adler = adler32_str(adler, vitals->unique_id);
            adler = adler32_str(adler, vitals->disk_model);
            adler = adler32_str(adler, vitals->disk_vendor);
            adler = adler32_str(adler, vitals->disk_probe_module);
            adler = adler32_str(adler, vitals->disk_probe_name);
            break;
        case DEV_TYPE_DM_MPATH:
        case DEV_TYPE_DM_RAID:
            adler = adler32_str(adler, vitals->dmname);
            break;
        case DEV_TYPE_PARTITION:
            /* XXX FIXME */
            break;
        case DEV_TYPE_LVM2_PV:
            adler = adler32_str(adler, vitals->unique_id);
            break;
        case DEV_TYPE_LVM2_VG:
            adler = adler32_str(adler, vitals->unique_id);
            adler = adler32_str(adler, vitals->vg_name);
            break;
        case DEV_TYPE_LVM2_LV:
            adler = adler32_str(adler, vitals->unique_id);
            adler = adler32_str(adler, vitals->lv_name);
            break;
        case DEV_TYPE_FS:
            adler = adler32_str(adler, vitals->unique_id);
            adler = adler32_str(adler, vitals->fs_type);
            adler = adler32_str(adler, vitals->fs_label);
            /* not including "fs_options" here, since we can't probe it
             * from the disk */
            break;
        case DEV_TYPE_MD_RAID:
            /* XXX FIXME */
            break;
        case DEV_TYPE_DM_ERROR:
            break;
        case DEV_TYPE_DM:
        case DEV_TYPE_FLOPPY:
        case DEV_TYPE_NONE:
        case DEV_TYPE_TREE:
            /* these are all "fake" or incomplete devices */
            return 0;
    }
    return adler;
}

#define xstrcmp(s0,s1) (s0 ? (s1 ? strcmp(s0,s1) : -1) : (s1 ? 1 : 0))

int nash_bdev_vitals_get_type(struct nash_bdev_vitals *vitals)
{
    return vitals->type;
}

int nash_bdev_vitals_type_cmp(struct nash_bdev_vitals *v0,
    struct nash_bdev_vitals *v1)
{
    int ret;
    if (v0 == v1)
        return 0;

    /* XXX ignoring everything but disks for now */
    if (v0->type != v1->type)
        return v1->type - v0->type;

    switch (v0->type) {
        case DEV_TYPE_DISK:
            ret = xstrcmp(v0->disk_probe_name, v1->disk_probe_name);
            if (ret)
                return ret;
            ret = xstrcmp(v0->disk_probe_module, v1->disk_probe_module);
            if (ret)
                return ret;
            break;
        default:
            /* XXX ignoring everything but disks for now */
            return -1;
    }
    return 0;
}

static int nash_bdev_vitals_manual_cmp(struct nash_bdev_vitals *v0,
        struct nash_bdev_vitals *v1)
{
    int ret = 0;

    if (!v0 || !v1)
        return v0 ? -1 : (v1 ? 1 : -1 );

    if (v0 == v1)
        return 0;

    if (v1->type != v0->type)
        return v1->type - v0->type;

    switch (v0->type) {
        case DEV_TYPE_TREE:
            return -1;
        case DEV_TYPE_DISK:
            if ((ret = xstrcmp(v0->unique_id, v1->unique_id)))
                break;
            if ((ret = xstrcmp(v0->disk_vendor, v1->disk_vendor)))
                break;
            if ((ret = xstrcmp(v0->disk_model, v1->disk_model)))
                break;
            if ((ret = xstrcmp(v0->disk_probe_module, v1->disk_probe_module)))
                break;
            ret = xstrcmp(v0->disk_probe_name, v1->disk_probe_name);
            break;
        case DEV_TYPE_DM_MPATH:
        case DEV_TYPE_DM_RAID:
            ret = xstrcmp(v0->dmname, v1->dmname);
            break;
        case DEV_TYPE_PARTITION:
            /* XXX FIXME */
            ret = -1;
            break;
        case DEV_TYPE_LVM2_PV:
            ret = xstrcmp(v0->unique_id, v1->unique_id);
            break;
        case DEV_TYPE_LVM2_VG:
            if ((ret = xstrcmp(v0->unique_id, v1->unique_id)))
                break;
            ret = xstrcmp(v0->vg_name, v1->vg_name);
            break;
        case DEV_TYPE_LVM2_LV:
            if ((ret = xstrcmp(v0->unique_id, v1->unique_id)))
                break;
            ret = xstrcmp(v0->lv_name, v1->lv_name);
            break;
        case DEV_TYPE_MD_RAID:
            /* XXX FIXME */
            ret = -1;
            break;
        case DEV_TYPE_FS:
            if ((ret = xstrcmp(v0->unique_id, v1->unique_id)))
                break;
            if ((ret = xstrcmp(v0->fs_label, v1->fs_label)))
                break;
            ret = xstrcmp(v0->fs_type, v1->fs_type);
            break;
        case DEV_TYPE_NONE:
        default:
            return -1;
            break;
    }
    return ret;
}

static void nash_bdev_vitals_hash_update(nashContext *nc,
        struct nash_bdev_vitals *vitals);

int nash_bdev_vitals_cmp(nashContext *nc, struct nash_bdev_vitals *v0,
    struct nash_bdev_vitals *v1)
{
    uint32_t hash0, hash1;

    if (!v0 || !v1)
        return v0 ? -1 : (v1 ? 1 : -1 );

    if (v0 == v1)
        return 0;

    if (v1->type != v0->type)
        return v1->type - v0->type;

    if (!v0->hash)
        nash_bdev_vitals_hash_update(nc, v0);
    hash0 = v0->hash;
    if (!v1->hash)
        nash_bdev_vitals_hash_update(nc, v1);
    hash1 = v1->hash;

    if (!hash0 || !hash1)
        return hash0 ? -1 : (hash1 ? 1 : -1 );

    return hash1 - hash0;
}

void nash_bdev_vitals_free(nashContext *nc, struct nash_bdev_vitals *vitals)
{
    if (!vitals || --vitals->refcount)
        return;

    nash_vitals_cache_remove(nc, vitals);
    xfree(vitals->name);

    switch (vitals->type) {
        case DEV_TYPE_TREE:
            break;
        case DEV_TYPE_DISK:
            xfree(vitals->unique_id);
            xfree(vitals->disk_device_node);
            xfree(vitals->disk_vendor);
            xfree(vitals->disk_model);
            xfree(vitals->disk_probe_module);
            xfree(vitals->disk_probe_name);
            break;
        case DEV_TYPE_DM_MPATH:
        case DEV_TYPE_DM_RAID:
            xfree(vitals->dmname);
            break;
        case DEV_TYPE_PARTITION:
            break;
        case DEV_TYPE_LVM2_PV:
            xfree(vitals->unique_id);
            break;
        case DEV_TYPE_LVM2_VG:
            xfree(vitals->unique_id);
            xfree(vitals->vg_name);
            break;
        case DEV_TYPE_LVM2_LV:
            xfree(vitals->unique_id);
            xfree(vitals->lv_name);
            break;
        case DEV_TYPE_MD_RAID:
            /* XXX fixme */
            break;
        case DEV_TYPE_FS:
            xfree(vitals->unique_id);
            xfree(vitals->fs_type);
            xfree(vitals->fs_label);
            xfree(vitals->fs_options);
            break;
        case DEV_TYPE_NONE:
        default:
            break;
    }
    free(vitals);
}

void nash_vitals_decref(nashContext *nc, struct nash_bdev_vitals *vitals)
    __attribute__((alias("nash_bdev_vitals_free")));

void nash_vitals_incref(struct nash_bdev_vitals *vitals)
{
    vitals->refcount++;
}

struct nash_bdev_vitals *nash_bdev_vitals_alloc(void)
{
    struct nash_bdev_vitals *vitals;

    if (!(vitals = calloc(1, sizeof (*vitals))))
        return NULL;

    vitals->type = DEV_TYPE_NONE;
    vitals->refcount = 1;
    return vitals;
}

struct nash_bdev_vitals *nash_bdev_vitals_from_blkent(nashContext *nc,
        struct blkent *blkent)
{
    struct nash_bdev_vitals *vitals, *new;

    if (!(vitals = nash_bdev_vitals_alloc()))
        return NULL;

    if (!blkent) {
        vitals->type = DEV_TYPE_NONE;
        return vitals;
    }

    vitals->type = nash_dev_node_get_type_from_blkent(blkent);
    if (vitals->type == DEV_TYPE_NONE)
        goto err;

    if (!(vitals->name = strdup(blkent->blk_name)))
        goto err;

    /* XXX need to check if it exists first? */
    switch (vitals->type) {
        case DEV_TYPE_DISK:
            vitals->unique_id = dupblkopt(blkent->blk_opts, "unique_id");
            vitals->disk_device_node =
                dupblkopt(blkent->blk_opts, "device_node");
            vitals->disk_vendor = dupblkopt(blkent->blk_opts, "vendor");
            vitals->disk_model = dupblkopt(blkent->blk_opts, "model");
            vitals->disk_probe_module =
                dupblkopt(blkent->blk_opts, "module");
            vitals->disk_probe_name =
                dupblkopt(blkent->blk_opts, "probe");
            if (!vitals->unique_id ||
                    !vitals->disk_device_node ||
                    !vitals->disk_vendor ||
                    !vitals->disk_model ||
                    !vitals->disk_probe_module ||
                    !vitals->disk_probe_name)
                goto err;
            break;
        case DEV_TYPE_DM_MPATH:
        case DEV_TYPE_DM_RAID:
            if (!(vitals->dmname = dupblkopt(blkent->blk_opts, "dmname")))
                goto err;
            break;
        case DEV_TYPE_PARTITION:
            /* FIXME: not even sure we really care about partitions... */
            break;
        case DEV_TYPE_LVM2_PV:
            if (!(vitals->unique_id = dupblkopt(blkent->blk_opts, "unique_id")))
                goto err;
            break;
        case DEV_TYPE_LVM2_VG:
            if (!(vitals->unique_id = dupblkopt(blkent->blk_opts, "unique_id")))
                goto err;
            if (!(vitals->vg_name = dupblkopt(blkent->blk_opts, "name")))
                goto err;
            break;
        case DEV_TYPE_LVM2_LV:
            if (!(vitals->unique_id = dupblkopt(blkent->blk_opts, "unique_id")))
                goto err;
            if (!(vitals->lv_name = dupblkopt(blkent->blk_opts, "name")))
                goto err;
            break;
        case DEV_TYPE_MD_RAID:
            /* FIXME */
            break;
        case DEV_TYPE_FS:
            if (!(vitals->unique_id = dupblkopt(blkent->blk_opts, "unique_id")))
                goto err;
            if (!(vitals->fs_type = dupblkopt(blkent->blk_opts, "type"))) 
                goto err;
            if (!(vitals->fs_label = dupblkopt(blkent->blk_opts, "label"))) 
                goto err;
            if (!(vitals->fs_options = dupblkopt(blkent->blk_opts, "options"))) 
                goto err;
            break;
        case DEV_TYPE_NONE:
            break;
    }

    if (vitals->type == DEV_TYPE_DISK && !vitals->unique_id)
        goto err;

    new = nash_vitals_cache_insert(nc, vitals);
    if (new != vitals) {
        if (new)
            nash_vitals_incref(new);
        nash_vitals_decref(nc, vitals);
    }
    return new;
err:
    nash_vitals_decref(nc, vitals);
    return NULL;
}

/*
 * this structure describes private probe data that relates to the probe
 * function rather than a device, so we only have to initialize probing once.
 */

struct probe_info {
    int id;
    char *name;
    void *priv;
    int (*alloc_priv)(struct probe_info *);
    void (*free_priv)(struct probe_info *);
    int (*do_probe)(nashContext *nc, struct probe_info *pi,
            struct nash_list *nodes, struct nash_bdev_vitals **vitalsp);
};

/* below this: probe functions */
/* PROBE_BDEVID prober */

static void probe_bdevid_destroy(struct probe_info *pi)
{
    struct bdevid *bdevid = pi->priv;
    if (!bdevid)
        return;

    bdevid_destroy(bdevid);
    pi->priv = NULL;
}

static int probe_bdevid_init(struct probe_info *pi)
{
    struct bdevid *bdevid;

    pi->priv = NULL;
    if (!(bdevid = bdevid_new("BDEVID_MODULE_PATH")))
        return -1;

    if (bdevid_module_load(bdevid, pi->name) < 0) {
        bdevid_destroy(bdevid);
        return -1;
    }

    pi->priv = bdevid;
    return 0;
}

struct bdevid_probe_priv {
    nashContext *nc;
    struct nash_dev_node *node;
    struct nash_bdev_vitals *vitals;
};

static int
bdevid_probe_visitor(struct bdevid_probe_result *result, void *privp)
{
    struct nash_bdev_vitals *vitals;
    struct bdevid_probe_priv *priv = (struct bdevid_probe_priv *)privp;
    char *attrs[] = {"module", "probe", "vendor", "model", "unique_id", NULL};
    int i, errnum;

    vitals = priv->vitals;
    vitals->type = DEV_TYPE_DISK;
    if (!(vitals->disk_device_node = strdup(priv->node->bdev->dev_path)))
        return -1;

    for (i = 0; attrs[i]; i++) {
        const char *key = attrs[i];
        const char *value = NULL;
        char *s = NULL;

        if (!(value = bdevid_pr_getattr(result, key)))
            continue;
        if (!(s = strdup(value)))
            goto err;
        if (!strcmp(key, "module"))
            vitals->disk_probe_module = s;
        else if (!strcmp(key, "probe"))
            vitals->disk_probe_name = s;
        else if (!strcmp(key, "vendor"))
            vitals->disk_vendor = s;
        else if (!strcmp(key, "model"))
            vitals->disk_model = s;
        else if (!strcmp(key, "unique_id"))
            vitals->unique_id = s;
        else if (s)
            free(s);
    }

    return 0;
err: 
    save_errno(errnum,
        xfree(vitals->disk_device_node);
        xfree(vitals->disk_probe_module);
        xfree(vitals->disk_probe_name);
        xfree(vitals->disk_vendor);
        xfree(vitals->disk_model);
        xfree(vitals->unique_id));

    return -1;
}

int probe_bdevid(nashContext *nc, struct probe_info *pi,
        struct nash_list *nodes, struct nash_bdev_vitals **vitalsp)
{
    struct bdevid_probe_priv priv = { .nc = nc };
    struct bdevid *bdevid;
    struct nash_bdev_vitals *vitals;
    struct nash_dev_node *node;
    int ret, errnum;

    if (!(bdevid = pi->priv))
        return -1;

    if (!(node = priv.node = nash_list_get(nodes, 0)))
        return -1;

    if (!(vitals = priv.vitals = nash_bdev_vitals_alloc()))
        return -1;

    if ((ret = bdevid_probe(bdevid, node->bdev->dev_path,
            bdevid_probe_visitor, &priv)) < 0) {
        save_errno(errnum, nash_bdev_vitals_free(nc, vitals));
        return -1;
    }
    if (vitals->type == DEV_TYPE_NONE) {
        nash_vitals_decref(nc, vitals);
        return -1;
    }

    if (vitals->type == DEV_TYPE_DISK && vitals->unique_id == NULL) {
        nash_vitals_decref(nc, vitals);
        return -1;
    }

    if (!(vitals->name = strdup(strrchr(vitals->disk_device_node, '/')+1))) {
        save_errno(errnum, nash_vitals_decref(nc, vitals));
        return -1;
    }

    *vitalsp = nash_vitals_cache_insert(nc, vitals);
    if (*vitalsp != vitals) {
        if (*vitalsp)
            nash_vitals_incref(*vitalsp);
        nash_vitals_decref(nc, vitals);
        priv.vitals = *vitalsp;
    }

    return ret;
}

/* end DEV_TYPE_DISK probe */

/* begin DEV_TYPE_DM_MPATH probe */
/* end DEV_TYPE_DM_MPATH probe */

/* begin DEV_TYPE_PARTITION probe */
/* end DEV_TYPE_PARTITION probe */

/*
 * this kindof sucks, but each probe has an ID (which can have only one bit
 * set), which are used for two purposes:
 * 1) to mark that we've done a probe we won't need to do again in
 *    node->probe_mask (not always set - dmraid for example can
 *    be used more than once, but there's still a 'final' time)
 * 2) to identify probe initialization data in probe_info[];
 */
#define PROBE_NONE          0x1 /* used for the end of the probe_info list */
#define PROBE_BDEVID_SCSI   0x2
#define PROBE_BDEVID_ATA    0x4
#define PROBE_BDEVID_USB    0x8
#define PROBE_PARTITION     0x10
#define PROBE_DM_MPATH      0x20 /* XXX might not exist... */
#define PROBE_DM_RAID       0x40
#define PROBE_MD_RAID       0x80
#define PROBE_LVM2_PV       0x100
#define PROBE_LVM2_VG       0x200
#define PROBE_LVM2_LV       0x400
#define PROBE_FS            0x800

static struct probe_info probe_info[] = {
    {
        .id = PROBE_BDEVID_SCSI,
        .name = "scsi",
        .alloc_priv = probe_bdevid_init,
        .free_priv = probe_bdevid_destroy,
        .do_probe = probe_bdevid,
    },
    {
        .id = PROBE_BDEVID_ATA,
        .name = "ata",
        .alloc_priv = probe_bdevid_init,
        .free_priv = probe_bdevid_destroy,
        .do_probe = probe_bdevid,
    },
    {
        .id = PROBE_BDEVID_USB,
        .name = "usb",
        .alloc_priv = probe_bdevid_init,
        .free_priv = probe_bdevid_destroy,
        .do_probe = probe_bdevid,
    },
#if 0
    {
        .id = PROBE_PARTITION,
        .name = "partition",
        .do_probe = probe_partitions,
    },
    /* XXX add probes for DM_MPATH, DM_RAID, MD_RAID,
     * LVM2_PV, LVM2_VG, and LVM2_LV
     */
    {
        .id = PROBE_FS,
        .name = "fs",
        .alloc_priv = probe_fs_init,
        .free_priv = probe_fs_destroy,
        .do_probe = probe_blkid,
    },
#endif
    { .id = PROBE_NONE, .name = "end of probe list" }
};

static struct probe_info *get_probe_info(int id)
{
    struct probe_info *pi = &probe_info[0];

    while (pi->id != PROBE_NONE && pi->id != id) 
        pi++;
    if (pi->id == PROBE_NONE)
        return NULL;
    return pi;
}

static int probes_initialized = 0;

int nash_vitals_initialize_probes(void)
{
    struct probe_info *pi;

    if (probes_initialized != 0)
        return probes_initialized;

    for (pi = &probe_info[0]; pi->id != PROBE_NONE; pi++) {
        /* we actually intentionally ignore failure here... maybe we shouldn't,
         * I can see both ways... */
        if (pi->alloc_priv)
            pi->alloc_priv(pi);
    }
    probes_initialized = 1;
    return 1;
}

void nash_vitals_destroy_probes(void)
{
    struct probe_info *pi;

    if (!probes_initialized)
        return;

    for (pi = &probe_info[0]; pi->id != PROBE_NONE; pi++) {
        if (pi->free_priv)
            pi->free_priv(pi);
    }
    probes_initialized = 0;
}

#ifdef DO_PRINT_DEV_TREE
int print_vitals(struct nash_bdev_vitals *vitals)
{
    if (vitals == NULL) {
        fprintf(stderr, "got NULL vitals\n");
        return 0;
    }
    switch (vitals->type) {
        case DEV_TYPE_FS:
            fprintf(stderr,"  vitals: (DEV_TYPE_FS)\n");
            fprintf(stderr,"    unique_id: %s type: %s label: %s\n",
                vitals->unique_id, vitals->fs_type, vitals->fs_label);
            return 1;
        case DEV_TYPE_LVM2_LV:
            fprintf(stderr, "  vitals: (DEV_TYPE_LVM2_LV)\n");
            fprintf(stderr, "    unique_id: %s lv_name: %s\n",
                vitals->unique_id, vitals->lv_name);
            return 1;
        case DEV_TYPE_LVM2_VG:
            fprintf(stderr, "  vitals: (DEV_TYPE_LVM2_VG)\n");
            fprintf(stderr, "    unique_id: %s vg_name: %s\n",
                vitals->unique_id, vitals->vg_name);
            return 1;
        case DEV_TYPE_LVM2_PV:
            fprintf(stderr, "  vitals: (DEV_TYPE_LVM2_PV)\n");
            fprintf(stderr, "    unique_id: %s\n", vitals->unique_id);
            return 1;
        case DEV_TYPE_DISK:
            fprintf(stderr, "  vitals: (DEV_TYPE_DISK)\n");
            fprintf(stderr, "    %s %s %s\n", vitals->disk_vendor, vitals->disk_model, vitals->unique_id);
                return 1;
        default:
            break;
    }
    return 0;
}
#else
int print_vitals(struct nash_bdev_vitals *vitals)
{
    return 1;
}
#endif

struct nash_list *nash_vitals_probe(nashContext *nc,
        struct nash_dev_node *node)
{
    struct probe_info *pi;
    struct nash_list *ret;
    struct nash_bdev_vitals *vitals = NULL;
    struct nash_list *nodes;

    if (node->flags & DEV_FLAG_BANISHED)
        return NULL;

    switch (node->type) {
        case DEV_TYPE_TREE:
        case DEV_TYPE_NONE:
            return NULL;
        default:
            if (!(ret = nash_ptr_list_new()))
                return NULL;
            break;
    }

    if (!(nodes = nash_ptr_list_new())) {
        int errnum;
        save_errno(errnum, nash_list_destroy(ret));
        return NULL;
    }

    switch (node->type) {
        case DEV_TYPE_TREE:
            /* this is here just for gcc's benefit. */
            break;
        /* this kindof sucks; it means people can't add modules themselves...*/
        case DEV_TYPE_DISK:
            nash_list_add(nodes, node);
            if (!(node->probe_mask & PROBE_BDEVID_SCSI)) {
                pi = get_probe_info(PROBE_BDEVID_SCSI);
                vitals = NULL;
                if (pi->do_probe(nc, pi, nodes, &vitals) >= 0) {
                    if (vitals) {
                        nash_list_add(ret, vitals);
                        print_vitals(vitals);
                    }
                    node->probe_mask |= PROBE_BDEVID_SCSI;
                }
            }
            if (!(node->probe_mask & PROBE_BDEVID_ATA)) {
                pi = get_probe_info(PROBE_BDEVID_ATA);
                vitals = NULL;
                if (pi->do_probe(nc, pi, nodes, &vitals) >= 0) {
                    if (vitals) {
                        nash_list_add(ret, vitals);
                        print_vitals(vitals);
                    }
                    node->probe_mask |= PROBE_BDEVID_ATA;
                }
            }
            if (!(node->probe_mask & PROBE_BDEVID_USB)) {
                pi = get_probe_info(PROBE_BDEVID_USB);
                vitals = NULL;
                if (pi->do_probe(nc, pi, nodes, &vitals) >= 0) {
                    if (vitals) {
                        nash_list_add(ret, vitals);
                        print_vitals(vitals);
                    }
                    node->probe_mask |= PROBE_BDEVID_USB;
                }
            }
            break;
        default:
        case DEV_TYPE_NONE:
            /* this is here just for gcc's benefit. */
            break;
    }

    while (nash_list_remove(nodes, 0) >= 0)
        ;
    nash_list_destroy(nodes);

    return ret;
}

int nash_dev_node_vitals_get(nashContext *nc,
    struct nash_dev_node *node, char *key, int i, char **value)
{
    int rc = -1;
    struct nash_bdev_vitals *vitals = nash_list_get(node->vitals, i);

    if (!vitals)
        return -1;

    if (!strcmp(key, "label")) {
        char *label = NULL;
        switch (vitals->type) {
            case DEV_TYPE_FS:
                label = vitals->fs_label;
                break;
            case DEV_TYPE_LVM2_LV:
                label = vitals->lv_name;
                break;
            case DEV_TYPE_LVM2_VG:
                label = vitals->vg_name;
                break;
            case DEV_TYPE_DM_MPATH:
            case DEV_TYPE_DM_RAID:
                label = vitals->dmname;
                break;
            default:
                rc = 0;
        }
        if (label) {
            *value = label;
            rc = 1;
        } else {
            rc = 0;
        }
    } else if (!strcmp(key, "unique_id")) {
        switch (vitals->type) {
            case DEV_TYPE_DISK:
            case DEV_TYPE_LVM2_PV:
            case DEV_TYPE_LVM2_VG:
            case DEV_TYPE_LVM2_LV:
            case DEV_TYPE_FS:
                *value = vitals->unique_id;
                rc = 1;
                break;
            default:
                rc = 0;
                break;
        }
    }
    return rc;
}

int nash_dev_node_vitals_has(nashContext *nc,
    struct nash_dev_node *node, char *key, char *value)
{
    char *name = NULL;
    int i = -1;
    int rc = 0;

    while (nash_dev_node_vitals_get(nc, node, key, ++i, &name) >= 0) {
        if (!name)
            continue;
        if (!strcmp(name, value)) {
            rc = 1;
            break;
        }
        name = NULL;
    }
    return rc;
}

struct nash_vitals_cache {
    struct nash_dict *dict;
};

static void nash_bdev_vitals_hash_update(nashContext *nc,
        struct nash_bdev_vitals *vitals)
{
    struct nash_dict *dict = nc->vitals->dict;
    int ret;
    uint32_t hash = 0;

    hash = nash_bdev_vitals_hash(vitals);
    ret = nash_dict_update(dict, vitals->hash, hash, vitals, vitals);
    vitals->hash = hash;
}

struct nash_bdev_vitals *nash_vitals_cache_search(nashContext *nc,
        struct nash_bdev_vitals *vitals)
{
    struct nash_dict *dict = nc->vitals->dict;    

    return nash_dict_get(dict, vitals);
}

void nash_vitals_cache_remove(nashContext *nc, struct nash_bdev_vitals *vitals)
{
    struct nash_dict *dict = nc->vitals->dict;
    struct nash_bdev_vitals *old;

    if (!(old = nash_dict_get(dict, vitals)))
        return;
    nash_dict_del(dict, old);
}

struct nash_bdev_vitals *nash_vitals_cache_insert(nashContext *nc,
        struct nash_bdev_vitals *vitals)
{
    struct nash_dict *dict = nc->vitals->dict;    
    struct nash_bdev_vitals *old;
    int ret;

    if (!(old = nash_dict_get(dict, vitals))) {
        if ((ret = nash_dict_add(dict, vitals, vitals)) < 0) {
            return NULL;
        }
        return vitals;
    }

    return old;
}

static int nash_vitals_cache_cmp(const void *a, const void *b)
{
    struct nash_bdev_vitals *v0 = *(struct nash_bdev_vitals **)a;
    struct nash_bdev_vitals *v1 = *(struct nash_bdev_vitals **)b;
    int ret;

    if (!v0 || !v1)
        return v0 ? -1 : (v1 ? 1 : -1 );

    if ((ret = nash_bdev_vitals_type_cmp(v0, v1)))
        return ret;

    return nash_bdev_vitals_manual_cmp(v0, v1);
}

static void nash_vitals_cache_destroy_helper(
        struct nash_vitals_cache **vitalsp)
{
    struct nash_vitals_cache *vitals = NULL;

    if (vitalsp) {
        vitals = *vitalsp;
        *vitalsp = NULL;
    }
    if (!vitals)
        return;

    nash_dict_free(vitals->dict);
    free(vitals);
}

int nash_vitals_cache_init(nashContext *nc)
{
    struct nash_vitals_cache *vitals;

    if (!(vitals = calloc(1, sizeof (*vitals))))
        return -1;

    if (!(vitals->dict = nash_dict_new(nash_bdev_vitals_hash,
                    nash_vitals_cache_cmp))) {
        int errnum;
        save_errno(errnum, nash_vitals_cache_destroy_helper(&vitals));
        return -1;
    }

    nc->vitals = vitals;

    return 0;
}

void nash_vitals_cache_destroy(nashContext *nc)
{
    nash_vitals_cache_destroy_helper(&nc->vitals);
}

/*
 * vim:ts=8:sw=4:sts=4:et
 */
